# Copyright (C) 2008 Robert C. Kirby (Texas Tech University)
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by David A. Ham (david.ham@imperial.ac.uk), 2014
# Modified by Lizao Li (lzlarryli@gmail.com), 2016

"""
Abstract class and particular implementations of finite element
reference simplex geometry/topology.

Provides an abstract base class and particular implementations for the
reference simplex geometry and topology.
The rest of FIAT is abstracted over this module so that different
reference element geometry (e.g. a vertex at (0,0) versus at (-1,-1))
and orderings of entities have a single point of entry.

Currently implemented are UFC and Default Line, Triangle and Tetrahedron.
"""
import operator
from collections import defaultdict
from functools import reduce
from itertools import chain, count, product
from math import factorial

import numpy
from gem.utils import safe_repr
from recursivenodes.nodes import _decode_family, _recursive

from FIAT.orientation_utils import (
    Orientation,
    make_cell_orientation_reflection_map_simplex,
    make_cell_orientation_reflection_map_tensorproduct,
    make_entity_permutations_simplex,
)

POINT = 0
LINE = 1
TRIANGLE = 2
TETRAHEDRON = 3
QUADRILATERAL = 11
HEXAHEDRON = 111
TENSORPRODUCT = 99

hypercube_shapes = {2: QUADRILATERAL, 3: HEXAHEDRON}


def multiindex_equal(d, isum, imin=0):
    """A generator for d-tuple multi-indices whose sum is isum and minimum is imin.
    """
    if d <= 0:
        return
    imax = isum - (d - 1) * imin
    if imax < imin:
        return
    for i in range(imin, imax):
        for a in multiindex_equal(d - 1, isum - i, imin=imin):
            yield a + (i,)
    yield (imin,) * (d - 1) + (imax,)


def lattice_iter(start, finish, depth):
    """Generator iterating over the depth-dimensional lattice of
    integers between start and (finish-1).  This works on simplices in
    0d, 1d, 2d, 3d, and beyond"""
    if depth == 0:
        yield tuple()
    elif depth == 1:
        for ii in range(start, finish):
            yield (ii,)
    else:
        for ii in range(start, finish):
            for jj in lattice_iter(start, finish - ii, depth - 1):
                yield jj + (ii,)


def make_lattice(verts, n, interior=0, variant=None):
    """Constructs a lattice of points on the simplex defined by verts.
    For example, the 1:st order lattice will be just the vertices.
    The optional argument interior specifies how many points from
    the boundary to omit.  For example, on a line with n = 2,
    and interior = 0, this function will return the vertices and
    midpoint, but with interior = 1, it will only return the
    midpoint."""
    if variant is None:
        variant = "equispaced"
    recursivenodes_families = {
        "equispaced": "equi",
        "equispaced_interior": "equi_interior",
        "gll": "lgl"}
    family = recursivenodes_families.get(variant, variant)
    family = _decode_family(family)
    D = len(verts)
    X = numpy.array(verts)
    get_point = lambda alpha: tuple(numpy.dot(_recursive(D - 1, n, alpha, family), X))
    return list(map(get_point, multiindex_equal(D, n, interior)))


def linalg_subspace_intersection(A, B):
    """Computes the intersection of the subspaces spanned by the
    columns of 2-dimensional arrays A,B using the algorithm found in
    Golub and van Loan (3rd ed) p. 604.  A should be in
    R^{m,p} and B should be in R^{m,q}.  Returns an orthonormal basis
    for the intersection of the spaces, stored in the columns of
    the result."""

    # check that vectors are in same space
    if A.shape[0] != B.shape[0]:
        raise Exception("Dimension error")

    # A,B are matrices of column vectors
    # compute the intersection of span(A) and span(B)

    # Compute the principal vectors/angles between the subspaces, G&vL
    # p.604
    (qa, _ra) = numpy.linalg.qr(A)
    (qb, _rb) = numpy.linalg.qr(B)

    C = numpy.dot(numpy.transpose(qa), qb)

    (y, c, _zt) = numpy.linalg.svd(C)

    U = numpy.dot(qa, y)

    rank_c = len([s for s in c if numpy.abs(1.0 - s) < 1.e-10])

    return U[:, :rank_c]


class Cell:
    """Abstract class for a reference cell.  Provides accessors for
    geometry (vertex coordinates) as well as topology (orderings of
    vertices that make up edges, faces, etc."""
    def __init__(self, shape, vertices, topology):
        """The constructor takes a shape code, the physical vertices expressed
        as a list of tuples of numbers, and the topology of a cell.

        The topology is stored as a dictionary of dictionaries t[i][j]
        where i is the dimension and j is the index of the facet of
        that dimension.  The result is a list of the vertices
        comprising the facet."""
        self.shape = shape
        self.vertices = vertices
        self.topology = topology

        # Given the topology, work out for each entity in the cell,
        # which other entities it contains.
        self.sub_entities = {}
        for dim, entities in topology.items():
            self.sub_entities[dim] = {}

            for e, v in entities.items():
                vertices = frozenset(v)
                sub_entities = []

                for dim_, entities_ in topology.items():
                    for e_, vertices_ in entities_.items():
                        if vertices.issuperset(vertices_):
                            sub_entities.append((dim_, e_))

                # Sort for the sake of determinism and by UFC conventions
                self.sub_entities[dim][e] = sorted(sub_entities)

        # Build super-entity dictionary by inverting the sub-entity dictionary
        self.super_entities = {dim: {entity: [] for entity in topology[dim]} for dim in topology}
        for dim0 in topology:
            for e0 in topology[dim0]:
                for dim1, e1 in self.sub_entities[dim0][e0]:
                    self.super_entities[dim1][e1].append((dim0, e0))

        # Build connectivity dictionary for easier queries
        self.connectivity = {}
        for dim0 in sorted(topology):
            for dim1 in sorted(topology):
                self.connectivity[(dim0, dim1)] = []

            for entity in sorted(topology[dim0]):
                children = self.sub_entities[dim0][entity]
                parents = self.super_entities[dim0][entity]
                for dim1 in sorted(topology):
                    neighbors = children if dim1 < dim0 else parents
                    d01_entities = tuple(e for d, e in neighbors if d == dim1)
                    self.connectivity[(dim0, dim1)].append(d01_entities)

        # Dictionary with derived cells
        self._split_cache = {}

    def __repr__(self):
        return f"{type(self).__name__}({self.shape!r}, {safe_repr(self.vertices)}, {self.topology!r})"

    def _key(self):
        """Hashable object key data (excluding type)."""
        # Default: only type matters
        return None

    def __hash__(self):
        return hash((type(self), self._key()))

    def get_shape(self):
        """Returns the code for the element's shape."""
        return self.shape

    def get_vertices(self):
        """Returns an iterable of the element's vertices, each stored as a
        tuple."""
        return self.vertices

    def get_spatial_dimension(self):
        """Returns the spatial dimension in which the element lives."""
        return len(self.vertices[0])

    def get_topology(self):
        """Returns a dictionary encoding the topology of the element.

        The dictionary's keys are the spatial dimensions (0, 1, ...)
        and each value is a dictionary mapping."""
        return self.topology

    def get_connectivity(self):
        """Returns a dictionary encoding the connectivity of the element.

        The dictionary's keys are the spatial dimensions pairs ((1, 0),
        (2, 0), (2, 1), ...) and each value is a list with entities
        of second dimension ordered by local dim0-dim1 numbering."""
        return self.connectivity

    def get_vertices_of_subcomplex(self, t):
        """Returns the tuple of vertex coordinates associated with the labels
        contained in the iterable t."""
        return tuple(self.vertices[ti] for ti in t)

    def get_dimension(self):
        """Returns the subelement dimension of the cell.  For tensor
        product cells, this a tuple of dimensions for each cell in the
        product.  For all other cells, this is the same as the spatial
        dimension."""
        raise NotImplementedError("Should be implemented in a subclass.")

    def construct_subelement(self, dimension):
        """Constructs the reference element of a cell subentity
        specified by subelement dimension.

        :arg dimension: `tuple` for tensor product cells, `int` otherwise
        """
        raise NotImplementedError("Should be implemented in a subclass.")

    def construct_subcomplex(self, dimension):
        """Constructs the reference subcomplex of the parent cell subentity
        specified by subcomplex dimension.

        :arg dimension: `tuple` for tensor product cells, `int` otherwise
        """
        if self.get_parent() is None:
            return self.construct_subelement(dimension)
        raise NotImplementedError("Should be implemented in a subclass.")

    def get_entity_transform(self, dim, entity_i):
        """Returns a mapping of point coordinates from the
        `entity_i`-th subentity of dimension `dim` to the cell.

        :arg dim: `tuple` for tensor product cells, `int` otherwise
        :arg entity_i: entity number (integer)
        """
        raise NotImplementedError("Should be implemented in a subclass.")

    def symmetry_group_size(self, dim):
        """Returns the size of the symmetry group of an entity of
        dimension `dim`."""
        raise NotImplementedError("Should be implemented in a subclass.")

    def cell_orientation_reflection_map(self):
        """Return the map indicating whether each possible cell orientation causes reflection (``1``) or not (``0``)."""
        raise NotImplementedError("Should be implemented in a subclass.")

    def extract_extrinsic_orientation(self, o):
        """Extract extrinsic orientation.

        Parameters
        ----------
        o : Orientation
            Total orientation.

        Returns
        -------
        Orientation
            Extrinsic orientation.

        """
        raise NotImplementedError("Should be implemented in a subclass.")

    def extract_intrinsic_orientation(self, o, axis):
        """Extract intrinsic orientation.

        Parameters
        ----------
        o : Orientation
            Total orientation.
        axis : int
            Reference cell axis for which intrinsic orientation is computed.

        Returns
        -------
        Orientation
            Intrinsic orientation.

        """
        raise NotImplementedError("Should be implemented in a subclass.")

    @property
    def extrinsic_orientation_permutation_map(self):
        """A map from extrinsic orientations to corresponding axis permutation matrices.

        Notes
        -----
        result[eo] gives the physical axis-reference axis permutation matrix corresponding to
        eo (extrinsic orientation).

        """
        raise NotImplementedError("Should be implemented in a subclass.")

    def is_simplex(self):
        return False

    def is_macrocell(self):
        return False

    def get_interior_facets(self, dim):
        """Return the interior facets this cell is a split and () otherwise."""
        return ()

    def get_parent(self):
        """Return the parent cell if this cell is a split and None otherwise."""
        return None

    def get_parent_complex(self):
        """Return the parent complex if this cell is a split and None otherwise."""
        return None

    def is_parent(self, other, strict=False):
        """Return whether this cell is the parent of the other cell."""
        parent = other
        if strict:
            parent = parent.get_parent_complex()
        while parent is not None:
            if self == parent:
                return True
            parent = parent.get_parent_complex()
        return False

    def __eq__(self, other):
        if self is other:
            return True
        A, B = self.get_vertices(), other.get_vertices()
        if not (len(A) == len(B) and numpy.allclose(A, B)):
            return False
        atop = self.get_topology()
        btop = other.get_topology()
        for dim in atop:
            if set(atop[dim].values()) != set(btop[dim].values()):
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __gt__(self, other):
        return other.is_parent(self, strict=True)

    def __lt__(self, other):
        return self.is_parent(other, strict=True)

    def __ge__(self, other):
        return other.is_parent(self, strict=False)

    def __le__(self, other):
        return self.is_parent(other, strict=False)


class SimplicialComplex(Cell):
    r"""Abstract class for a simplicial complex.

    This consists of list of vertex locations and a topology map defining facets.
    """
    def __init__(self, shape, vertices, topology):
        # Make sure that every facet has the right number of vertices to be
        # a simplex.
        for dim in topology:
            for entity in topology[dim]:
                assert len(topology[dim][entity]) == dim + 1

        super().__init__(shape, vertices, topology)

    def compute_normal(self, facet_i, cell=None):
        """Returns the unit normal vector to facet i of codimension 1."""

        t = self.get_topology()
        sd = self.get_spatial_dimension()

        # To handle simplicial complex case:
        # Find a subcell of which facet_i is on the boundary
        # Note: this is trivial and vastly overengineered for the single-cell
        # case.
        if cell is None:
            cell = next(k for k, facets in enumerate(self.connectivity[(sd, sd-1)])
                        if facet_i in facets)
        verts = numpy.asarray(self.get_vertices_of_subcomplex(t[sd][cell]))
        # Interval case
        if self.get_shape() == LINE:
            v_i = t[1][cell].index(t[0][facet_i][0])
            n = verts[v_i] - verts[[1, 0][v_i]]
            return n / numpy.linalg.norm(n)

        # vectors from vertex 0 to each other vertex.
        vert_vecs_from_v0 = verts[1:, :] - verts[:1, :]

        (u, s, _) = numpy.linalg.svd(vert_vecs_from_v0)
        rank = len([si for si in s if si > 1.e-10])

        # this is the set of vectors that span the simplex
        spanu = u[:, :rank]

        vert_coords_of_facet = \
            self.get_vertices_of_subcomplex(t[sd-1][facet_i])

        # now I find everything normal to the facet.
        vcf = numpy.asarray(vert_coords_of_facet)
        facet_span = vcf[1:, :] - vcf[:1, :]
        (_, sf, vft) = numpy.linalg.svd(facet_span)

        # now get the null space from vft
        rankfacet = len([si for si in sf if si > 1.e-10])
        facet_normal_space = numpy.transpose(vft[rankfacet:, :])

        # now, I have to compute the intersection of
        # facet_span with facet_normal_space
        foo = linalg_subspace_intersection(facet_normal_space, spanu)

        num_cols = foo.shape[1]

        if num_cols != 1:
            raise Exception("barf in normal computation")

        # now need to get the correct sign
        # get a vector in the direction
        nfoo = foo[:, 0]

        # what is the vertex not in the facet?
        verts_set = set(t[sd][cell])
        verts_facet = set(t[sd - 1][facet_i])
        verts_diff = verts_set.difference(verts_facet)
        if len(verts_diff) != 1:
            raise Exception("barf in normal computation: getting sign")
        vert_off = verts_diff.pop()
        vert_on = verts_facet.pop()

        # get a vector from the off vertex to the facet
        v_to_facet = numpy.array(self.vertices[vert_on]) \
            - numpy.array(self.vertices[vert_off])

        if numpy.dot(v_to_facet, nfoo) > 0.0:
            return nfoo
        else:
            return -nfoo

    def compute_tangents(self, dim, i):
        """Computes tangents in any dimension based on differences
        between vertices and the first vertex of the i:th facet
        of dimension dim.  Returns a (possibly empty) list.
        These tangents are *NOT* normalized to have unit length."""
        t = self.get_topology()
        vs = numpy.array(self.get_vertices_of_subcomplex(t[dim][i]))
        return vs[1:] - vs[:1]

    def compute_normalized_tangents(self, dim, i):
        """Computes tangents in any dimension based on differences
        between vertices and the first vertex of the i:th facet
        of dimension dim.  Returns a (possibly empty) list.
        These tangents are normalized to have unit length."""
        ts = self.compute_tangents(dim, i)
        ts /= numpy.linalg.norm(ts, axis=1)[:, None]
        return ts

    def compute_edge_tangent(self, edge_i):
        """Computes the nonnormalized tangent to a 1-dimensional facet.
        returns a single vector."""
        t = self.get_topology()
        vs = numpy.asarray(self.get_vertices_of_subcomplex(t[1][edge_i]))
        return vs[1] - vs[0]

    def compute_normalized_edge_tangent(self, edge_i):
        """Computes the unit tangent vector to a 1-dimensional facet"""
        v = self.compute_edge_tangent(edge_i)
        v /= numpy.linalg.norm(v)
        return v

    def compute_face_tangents(self, face_i):
        """Computes the two tangents to a face.  Only implemented
        for a tetrahedron."""
        if self.get_spatial_dimension() != 3:
            raise Exception("can't get face tangents yet")
        t = self.get_topology()
        vs = numpy.asarray(self.get_vertices_of_subcomplex(t[2][face_i]))
        return vs[1:] - vs[:1]

    def compute_face_edge_tangents(self, dim, entity_id):
        """Computes all the edge tangents of any k-face with k>=1.
        The result is a array of binom(dim+1,2) vectors.
        This agrees with `compute_edge_tangent` when dim=1.
        """
        vert_ids = self.get_topology()[dim][entity_id]
        vert_coords = numpy.asarray(self.get_vertices_of_subcomplex(vert_ids))
        v0 = []
        v1 = []
        for source in range(dim):
            for dest in range(source + 1, dim + 1):
                v0.append(source)
                v1.append(dest)
        return vert_coords[v1] - vert_coords[v0]

    def make_points(self, dim, entity_id, order, variant=None, interior=1):
        """Constructs a lattice of points on the entity_id:th
        facet of dimension dim.  Order indicates how many points to
        include in each direction."""
        if dim == 0:
            return (self.get_vertices()[entity_id], )
        elif 0 < dim <= self.get_spatial_dimension():
            entity_verts = \
                self.get_vertices_of_subcomplex(
                    self.get_topology()[dim][entity_id])
            return make_lattice(entity_verts, order, interior=interior, variant=variant)
        else:
            raise ValueError("illegal dimension")

    def volume(self):
        """Computes the volume of the simplicial complex in the appropriate
        dimensional measure."""
        sd = self.get_spatial_dimension()
        return sum(self.volume_of_subcomplex(sd, k)
                   for k in self.topology[sd])

    def volume_of_subcomplex(self, dim, facet_no):
        vids = self.topology[dim][facet_no]
        return volume(self.get_vertices_of_subcomplex(vids))

    def compute_scaled_normal(self, facet_i):
        """Returns the unit normal to facet_i of scaled by the
        volume of that facet."""
        dim = self.get_spatial_dimension()
        if dim == 2:
            n, = self.compute_tangents(dim-1, facet_i)
            n[0], n[1] = n[1], -n[0]
            return n
        elif dim == 3:
            return -numpy.cross(*self.compute_tangents(dim-1, facet_i))
        v = self.volume_of_subcomplex(dim - 1, facet_i)
        return self.compute_normal(facet_i) * v

    def compute_reference_normal(self, facet_dim, facet_i):
        """Returns the unit normal in infinity norm to facet_i."""
        assert facet_dim == self.get_spatial_dimension() - 1
        n = SimplicialComplex.compute_normal(self, facet_i)  # skip UFC overrides
        return n / numpy.linalg.norm(n, numpy.inf)

    def get_entity_transform(self, dim, entity):
        """Returns a mapping of point coordinates from the
        `entity`-th subentity of dimension `dim` to the cell.

        :arg dim: subentity dimension (integer)
        :arg entity: entity number (integer)
        """
        topology = self.get_topology()
        celldim = self.get_spatial_dimension()
        codim = celldim - dim
        if dim == 0:
            # Special case vertices.
            i, = topology[dim][entity]
            offset = numpy.asarray(self.get_vertices()[i])
            C = numpy.zeros((dim, ) + offset.shape)
        elif dim == celldim and len(self.topology[celldim]) == 1:
            assert entity == 0
            return lambda x: x
        else:
            subcell = self.construct_subelement(dim)
            subdim = subcell.get_spatial_dimension()
            assert subdim == celldim - codim

            # Entity vertices in entity space.
            v_e = numpy.asarray(subcell.get_vertices())
            A = v_e[1:] - v_e[:1]

            # Entity vertices in cell space.
            v_c = numpy.asarray(self.get_vertices_of_subcomplex(topology[dim][entity]))
            B = v_c[1:] - v_c[:1]

            C = numpy.linalg.solve(A, B)

            offset = v_c[0] - numpy.dot(v_e[0], C)

        def transform(point):
            out = numpy.dot(point, C)
            return numpy.add(out, offset, out=out)

        return transform

    def get_dimension(self):
        """Returns the subelement dimension of the cell.  Same as the
        spatial dimension."""
        return self.get_spatial_dimension()

    def compute_barycentric_coordinates(self, points, entity=None, rescale=False):
        """Returns the barycentric coordinates of a list of points on the complex."""
        if len(points) == 0:
            return points
        if entity is None:
            entity = (self.get_spatial_dimension(), 0)
        entity_dim, entity_id = entity
        top = self.get_topology()
        sd = self.get_spatial_dimension()

        # get a subcell containing the entity and the restriction indices of the entity
        indices = slice(None)
        subcomplex = top[entity_dim][entity_id]
        if entity_dim != sd:
            cell_id = self.connectivity[(entity_dim, sd)][entity_id][0]
            indices = [i for i, v in enumerate(top[sd][cell_id]) if v in subcomplex]
            subcomplex = top[sd][cell_id]

        cell_verts = self.get_vertices_of_subcomplex(subcomplex)
        ref_verts = numpy.eye(sd + 1)
        A, b = make_affine_mapping(cell_verts, ref_verts)
        A, b = A[indices], b[indices]
        if rescale:
            # rescale barycentric coordinates by the height wrt. to the facet
            h = 1 / numpy.linalg.norm(A, axis=1)
            b *= h
            A *= h[:, None]
        out = numpy.dot(points, A.T)
        return numpy.add(out, b, out=out)

    def compute_bubble(self, points, entity=None):
        """Returns the lowest-order bubble on an entity evaluated at the given
        points on the cell."""
        return numpy.prod(self.compute_barycentric_coordinates(points, entity), axis=1)

    def distance_to_point_l1(self, points, entity=None, rescale=False):
        # noqa: D301
        """Get the L1 distance (aka 'manhatten', 'taxicab' or rectilinear
        distance) from an entity to a point with 0.0 if the point is inside the entity.

        Parameters
        ----------
        points : numpy.ndarray or list
            The coordinates of the points.
        entity : tuple or None
            A tuple of entity dimension and entity id.
        rescale : bool
            If true, the L1 distance is measured with respect to rescaled
            barycentric coordinates, such that the L1 and L2 distances agree
            for points opposite to a single facet.

        Returns
        -------
        numpy.float64 or numpy.ndarray
            The L1 distance, also known as taxicab, manhatten or rectilinear
            distance, of the cell to the point. If 0.0 the point is inside the
            cell.

        Notes
        -----

        This is done with the help of barycentric coordinates where the general
        algorithm is to compute the most negative (i.e. minimum) barycentric
        coordinate then return its negative. For implementation reasons we
        return the sum of all the negative barycentric coordinates. In each of
        the below examples the point coordinate is `X` with appropriate
        dimensions.

        Consider, for example, a UFCInterval. We have two vertices which make
        the interval,
            `P0 = [0]` and
            `P1 = [1]`.
        Our point is
            `X = [x]`.
        Barycentric coordinates are defined as
            `X = alpha * P0 + beta * P1` where
            `alpha + beta = 1.0`.
        The solution is
            `alpha = 1 - X[0] = 1 - x` and
            `beta = X[0] = x`.
        If both `alpha` and `beta` are positive, the point is inside the
        reference interval.

        `---regionA---P0=0------P1=1---regionB---`

        If we are in `regionA`, `alpha` is negative and
        `-alpha = X[0] - 1.0` is the (positive) distance from `P0`.
        If we are in `regionB`, `beta` is negative and `-beta = -X[0]` is
        the exact (positive) distance from `P1`. Since we are in 1D the L1
        distance is the same as the L2 distance. If we are in the interval we
        can just return 0.0.

        Things get more complicated when we consider higher dimensions.
        Consider a UFCTriangle. We have three vertices which make the
        reference triangle,
            `P0 = (0, 0)`,
            `P1 = (1, 0)` and
            `P2 = (0, 1)`.
        Our point is
            `X = [x, y]`.
        Below is a diagram of the cell (which may not render correctly in
        sphinx):

        .. code-block:: text
        ```
                y-axis
                |
                |
          (0,1) P2
                | \\
                |  \\
                |   \\
                |    \\
                |  T  \\
                |      \\
                |       \\
                |        \\
            ---P0--------P1--- x-axis
          (0,0) |         (1,0)
        ```

        Barycentric coordinates are defined as
            `X = alpha * P0 + beta * P1 + gamma * P2` where
            `alpha + beta + gamma = 1.0`.
        The solution is
            `alpha = 1 - X[0] - X[1] = 1 - x - y`,
            `beta = X[0] = x` and
            `gamma = X[1] = y`.
        If all three are positive, the point is inside the reference cell.
        If any are negative, we are outside it. The absolute sum of any
        negative barycentric coordinates usefully gives the L1 distance from
        the cell to the point. For example the point (1,1) has L1 distance
        1 from the cell: on this case alpha = -1, beta = 1 and gamma = 1.
        -alpha = 1 is the L1 distance. For comparison the L2 distance (the
        length of the vector from the nearest point on the cell to the point)
        is sqrt(0.5^2 + 0.5^2) = 0.707. Similarly the point (-1.0, -1.0) has
        alpha = 3, beta = -1 and gamma = -1. The absolute sum of beta and gamma
        2 which is again the L1 distance. The L2 distance in this case is
        sqrt(1^2 + 1^2) = 1.414.

        For a UFCTetrahedron we have four vertices
            `P0 = (0,0,0)`,
            `P1 = (1,0,0)`,
            `P2 = (0,1,0)` and
            `P3 = (0,0,1)`.
        Our point is
            `X = [x, y, z]`.
        The barycentric coordinates are defined as
            `X = alpha * P0 + beta * P1 + gamma * P2 + delta * P3`
            where
            `alpha + beta + gamma + delta = 1.0`.
        The solution is
            `alpha = 1 - X[0] - X[1] - X[2] = 1 - x - y - z`,
            `beta = X[0] = x`,
            `gamma = X[1] = y` and
            `delta = X[2] = z`.
        The rules are the same as for the tetrahedron but with one extra
        barycentric coordinate. Our approximate distance, the absolute sum of
        the negative barycentric coordinates, is at worse around 4 times the
        actual distance to the tetrahedron.

        """
        # sum the negative part of each barycentric coordinate
        bary = self.compute_barycentric_coordinates(points, entity=entity, rescale=rescale)
        return 0.5 * abs(numpy.sum(abs(bary) - bary, axis=-1))

    def contains_point(self, point, epsilon=0.0, entity=None):
        """Checks if reference cell contains given point
        (with numerical tolerance as given by the L1 distance (aka 'manhatten',
        'taxicab' or rectilinear distance) to the cell.

        Parameters
        ----------
        point : numpy.ndarray, list or symbolic expression
            The coordinates of the point.
        epsilon : float
            The tolerance for the check.
        entity : tuple or None
            A tuple of entity dimension and entity id.

        Returns
        -------
        bool : True if the point is inside the cell, False otherwise.

        """
        return self.distance_to_point_l1(point, entity=entity) <= epsilon

    def extract_extrinsic_orientation(self, o):
        """Extract extrinsic orientation.

        Parameters
        ----------
        o : Orientation
            Total orientation.

        Returns
        -------
        Orientation
            Extrinsic orientation.

        """
        if not isinstance(o, Orientation):
            raise TypeError(f"Expecting an instance of Orientation : got {o}")
        return 0

    def extract_intrinsic_orientation(self, o, axis):
        """Extract intrinsic orientation.

        Parameters
        ----------
        o : Orientation
            Total orientation.
        axis : int
            Reference cell axis for which intrinsic orientation is computed.

        Returns
        -------
        Orientation
            Intrinsic orientation.

        """
        if not isinstance(o, Orientation):
            raise TypeError(f"Expecting an instance of Orientation : got {o}")
        if axis != 0:
            raise ValueError(f"axis ({axis}) != 0")
        return o

    @property
    def extrinsic_orientation_permutation_map(self):
        """A map from extrinsic orientations to corresponding axis permutation matrices.

        Notes
        -----
        result[eo] gives the physical axis-reference axis permutation matrix corresponding to
        eo (extrinsic orientation).

        """
        return numpy.diag((1, )).astype(int).reshape((1, 1, 1))


class Simplex(SimplicialComplex):
    r"""Abstract class for a reference simplex.

    Orientation of a physical cell is computed systematically
    by comparing the canonical orderings of its facets and
    the facets in the FIAT reference cell.

    As an example, we compute the orientation of a
    triangular cell:

       +                    +
       | \                  | \
       1   0               47   42
       |     \              |     \
       +--2---+             +--43--+
    FIAT canonical     Mapped example physical cell

    Suppose that the facets of the physical cell
    are canonically ordered as:

    C = [43, 42, 47]

    FIAT facet to Physical facet map is given by:

    M = [42, 47, 43]

    Then the orientation of the cell is computed as:

    C.index(M[0]) = 1; C.remove(M[0])
    C.index(M[1]) = 1; C.remove(M[1])
    C.index(M[2]) = 0; C.remove(M[2])

    o = (1 * 2!) + (1 * 1!) + (0 * 0!) = 3
    """
    def is_simplex(self):
        return True

    def symmetry_group_size(self, dim):
        return factorial(dim + 1)

    def cell_orientation_reflection_map(self):
        """Return the map indicating whether each possible cell orientation causes reflection (``1``) or not (``0``)."""
        return make_cell_orientation_reflection_map_simplex(self.get_dimension())

    def get_facet_element(self):
        dimension = self.get_spatial_dimension()
        return self.construct_subelement(dimension - 1)


# Backwards compatible name
ReferenceElement = Simplex


class UFCSimplex(Simplex):

    def construct_subelement(self, dimension):
        """Constructs the reference element of a cell subentity
        specified by subelement dimension.

        :arg dimension: subentity dimension (integer)
        """
        return ufc_simplex(dimension)


class DefaultSimplex(Simplex):

    def construct_subelement(self, dimension):
        """Constructs the reference element of a cell subentity
        specified by subelement dimension.

        :arg dimension: subentity dimension (integer)
        """
        return default_simplex(dimension)


class SymmetricSimplex(Simplex):

    def construct_subelement(self, dimension):
        """Constructs the reference element of a cell subentity
        specified by subelement dimension.

        :arg dimension: subentity dimension (integer)
        """
        return symmetric_simplex(dimension)


class Point(Simplex):
    """This is the reference point."""

    def __init__(self):
        verts = ((),)
        topology = {0: {0: (0,)}}
        super().__init__(POINT, verts, topology)

    def construct_subelement(self, dimension):
        """Constructs the reference element of a cell subentity
        specified by subelement dimension.

        :arg dimension: subentity dimension (integer). Must be zero.
        """
        assert dimension == 0
        return self


class DefaultLine(DefaultSimplex):
    """This is the reference line with vertices (-1.0,) and (1.0,)."""

    def __init__(self):
        verts = ((-1.0,), (1.0,))
        edges = {0: (0, 1)}
        topology = {0: {0: (0,), 1: (1,)},
                    1: edges}
        super().__init__(LINE, verts, topology)


class UFCInterval(UFCSimplex):
    """This is the reference interval with vertices (0.0,) and (1.0,)."""

    def __init__(self):
        verts = ((0.0,), (1.0,))
        edges = {0: (0, 1)}
        topology = {0: {0: (0,), 1: (1,)},
                    1: edges}
        super().__init__(LINE, verts, topology)


class DefaultTriangle(DefaultSimplex):
    """This is the reference triangle with vertices (-1.0,-1.0),
    (1.0,-1.0), and (-1.0,1.0)."""

    def __init__(self):
        verts = ((-1.0, -1.0), (1.0, -1.0), (-1.0, 1.0))
        edges = {0: (1, 2),
                 1: (2, 0),
                 2: (0, 1)}
        faces = {0: (0, 1, 2)}
        topology = {0: {0: (0,), 1: (1,), 2: (2,)},
                    1: edges, 2: faces}
        super().__init__(TRIANGLE, verts, topology)


class UFCTriangle(UFCSimplex):
    """This is the reference triangle with vertices (0.0,0.0),
    (1.0,0.0), and (0.0,1.0)."""

    def __init__(self):
        verts = ((0.0, 0.0), (1.0, 0.0), (0.0, 1.0))
        edges = {0: (1, 2), 1: (0, 2), 2: (0, 1)}
        faces = {0: (0, 1, 2)}
        topology = {0: {0: (0,), 1: (1,), 2: (2,)},
                    1: edges, 2: faces}
        super().__init__(TRIANGLE, verts, topology)

    def compute_normal(self, i):
        "UFC consistent normal"
        t = self.compute_tangents(1, i)[0]
        n = numpy.array((t[1], -t[0]))
        return n / numpy.linalg.norm(n)


class IntrepidTriangle(Simplex):
    """This is the Intrepid triangle with vertices (0,0),(1,0),(0,1)"""

    def __init__(self):
        verts = ((0.0, 0.0), (1.0, 0.0), (0.0, 1.0))
        edges = {0: (0, 1),
                 1: (1, 2),
                 2: (2, 0)}
        faces = {0: (0, 1, 2)}
        topology = {0: {0: (0,), 1: (1,), 2: (2,)},
                    1: edges, 2: faces}
        super().__init__(TRIANGLE, verts, topology)

    def get_facet_element(self):
        # I think the UFC interval is equivalent to what the
        # IntrepidInterval would be.
        return UFCInterval()


class DefaultTetrahedron(DefaultSimplex):
    """This is the reference tetrahedron with vertices (-1,-1,-1),
    (1,-1,-1),(-1,1,-1), and (-1,-1,1)."""

    def __init__(self):
        verts = ((-1.0, -1.0, -1.0), (1.0, -1.0, -1.0),
                 (-1.0, 1.0, -1.0), (-1.0, -1.0, 1.0))
        vs = {0: (0, ),
              1: (1, ),
              2: (2, ),
              3: (3, )}
        edges = {0: (1, 2),
                 1: (2, 0),
                 2: (0, 1),
                 3: (0, 3),
                 4: (1, 3),
                 5: (2, 3)}
        faces = {0: (1, 3, 2),
                 1: (2, 3, 0),
                 2: (3, 1, 0),
                 3: (0, 1, 2)}
        tets = {0: (0, 1, 2, 3)}
        topology = {0: vs, 1: edges, 2: faces, 3: tets}
        super().__init__(TETRAHEDRON, verts, topology)


class IntrepidTetrahedron(Simplex):
    """This is the reference tetrahedron with vertices (0,0,0),
    (1,0,0),(0,1,0), and (0,0,1) used in the Intrepid project."""

    def __init__(self):
        verts = ((0.0, 0.0, 0.0), (1.0, 0.0, 0.0), (0.0, 1.0, 0.0), (0.0, 0.0, 1.0))
        vs = {0: (0, ),
              1: (1, ),
              2: (2, ),
              3: (3, )}
        edges = {0: (0, 1),
                 1: (1, 2),
                 2: (2, 0),
                 3: (0, 3),
                 4: (1, 3),
                 5: (2, 3)}
        faces = {0: (0, 1, 3),
                 1: (1, 2, 3),
                 2: (0, 3, 2),
                 3: (0, 2, 1)}
        tets = {0: (0, 1, 2, 3)}
        topology = {0: vs, 1: edges, 2: faces, 3: tets}
        super().__init__(TETRAHEDRON, verts, topology)

    def get_facet_element(self):
        return IntrepidTriangle()


class UFCTetrahedron(UFCSimplex):
    """This is the reference tetrahedron with vertices (0,0,0),
    (1,0,0),(0,1,0), and (0,0,1)."""

    def __init__(self):
        verts = ((0.0, 0.0, 0.0), (1.0, 0.0, 0.0), (0.0, 1.0, 0.0), (0.0, 0.0, 1.0))
        vs = {0: (0, ),
              1: (1, ),
              2: (2, ),
              3: (3, )}
        edges = {0: (2, 3),
                 1: (1, 3),
                 2: (1, 2),
                 3: (0, 3),
                 4: (0, 2),
                 5: (0, 1)}
        faces = {0: (1, 2, 3),
                 1: (0, 2, 3),
                 2: (0, 1, 3),
                 3: (0, 1, 2)}
        tets = {0: (0, 1, 2, 3)}
        topology = {0: vs, 1: edges, 2: faces, 3: tets}
        super().__init__(TETRAHEDRON, verts, topology)

    def compute_normal(self, i):
        "UFC consistent normals."
        t = self.compute_tangents(2, i)
        n = numpy.cross(t[0], t[1])
        return -2.0 * n / numpy.linalg.norm(n)


class TensorProductCell(Cell):
    """A cell that is the product of FIAT cells."""

    def __init__(self, *cells):
        # Vertices
        vertices = tuple(tuple(chain(*coords))
                         for coords in product(*[cell.get_vertices()
                                                 for cell in cells]))

        # Topology
        shape = tuple(len(c.get_vertices()) for c in cells)
        topology = {}
        for dim in product(*[cell.get_topology().keys()
                             for cell in cells]):
            topology[dim] = {}
            topds = [cell.get_topology()[d]
                     for cell, d in zip(cells, dim)]
            for tuple_ei in product(*[sorted(topd)for topd in topds]):
                tuple_vs = list(product(*[topd[ei]
                                          for topd, ei in zip(topds, tuple_ei)]))
                vs = tuple(numpy.ravel_multi_index(numpy.transpose(tuple_vs), shape))
                topology[dim][tuple_ei] = vs
            # flatten entity numbers
            topology[dim] = dict(enumerate(topology[dim][key]
                                           for key in sorted(topology[dim])))

        super().__init__(TENSORPRODUCT, vertices, topology)
        self.cells = tuple(cells)

    def __repr__(self):
        return f"{type(self).__name__}({self.cells!r})"

    def _key(self):
        return self.cells

    @staticmethod
    def _split_slices(lengths):
        n = len(lengths)
        delimiter = [0] * (n + 1)
        for i in range(n):
            delimiter[i + 1] = delimiter[i] + lengths[i]
        return [slice(delimiter[i], delimiter[i+1])
                for i in range(n)]

    def get_dimension(self):
        """Returns the subelement dimension of the cell, a tuple of
        dimensions for each cell in the product."""
        return tuple(c.get_dimension() for c in self.cells)

    def construct_subelement(self, dimension):
        """Constructs the reference element of a cell subentity
        specified by subelement dimension.

        :arg dimension: dimension in each "direction" (tuple)
        """
        return TensorProductCell(*[c.construct_subelement(d)
                                   for c, d in zip(self.cells, dimension)])

    def construct_subcomplex(self, dimension):
        """Constructs the reference subcomplex of the parent cell subentity
        specified by subcomplex dimension.

        :arg dimension: dimension in each "direction" (tuple)
        """
        return TensorProductCell(*[c.construct_subcomplex(d)
                                   for c, d in zip(self.cells, dimension)])

    def get_entity_transform(self, dim, entity_i):
        """Returns a mapping of point coordinates from the
        `entity_i`-th subentity of dimension `dim` to the cell.

        :arg dim: subelement dimension (tuple)
        :arg entity_i: entity number (integer)
        """
        # unravel entity_i
        shape = tuple(len(c.get_topology()[d])
                      for c, d in zip(self.cells, dim))
        alpha = numpy.unravel_index(entity_i, shape)

        # entity transform on each subcell
        sct = [c.get_entity_transform(d, i)
               for c, d, i in zip(self.cells, dim, alpha)]

        slices = TensorProductCell._split_slices(dim)

        def transform(point):
            point = numpy.asarray(point)
            return numpy.concatenate(tuple(t(point[..., s])
                                     for t, s in zip(sct, slices)), axis=-1)

        return transform

    def volume(self):
        """Computes the volume in the appropriate dimensional measure."""
        return numpy.prod([c.volume() for c in self.cells])

    def compute_reference_normal(self, facet_dim, facet_i):
        """Returns the unit normal in infinity norm to facet_i of
        subelement dimension facet_dim."""
        assert len(facet_dim) == len(self.get_dimension())
        indicator = numpy.array(self.get_dimension()) - numpy.array(facet_dim)
        (cell_i,), = numpy.nonzero(indicator)

        n = []
        for i, c in enumerate(self.cells):
            if cell_i == i:
                n.extend(c.compute_reference_normal(facet_dim[i], facet_i))
            else:
                n.extend([0] * c.get_spatial_dimension())
        return numpy.asarray(n)

    def contains_point(self, point, epsilon=0.0):
        """Checks if reference cell contains given point
        (with numerical tolerance as given by the L1 distance (aka 'manhattan',
        'taxicab' or rectilinear distance) to the cell.

        Parameters
        ----------
        point : numpy.ndarray, list or symbolic expression
            The coordinates of the point.
        epsilon : float
            The tolerance for the check.

        Returns
        -------
        bool : True if the point is inside the cell, False otherwise.

        """
        subcell_dimensions = self.get_dimension()
        assert len(point) == sum(subcell_dimensions)
        point_slices = TensorProductCell._split_slices(subcell_dimensions)
        return reduce(operator.and_,
                      (c.contains_point(point[s], epsilon=epsilon)
                       for c, s in zip(self.cells, point_slices)),
                      True)

    def distance_to_point_l1(self, point, rescale=False):
        """Get the L1 distance (aka 'manhatten', 'taxicab' or rectilinear
        distance) to a point with 0.0 if the point is inside the cell.

        For more information see the docstring for the UFCSimplex method."""
        subcell_dimensions = self.get_dimension()
        assert len(point) == sum(subcell_dimensions)
        point_slices = TensorProductCell._split_slices(subcell_dimensions)
        point = numpy.asarray(point)
        return sum(c.distance_to_point_l1(point[..., s], rescale=rescale)
                   for c, s in zip(self.cells, point_slices))

    def symmetry_group_size(self, dim):
        return tuple(c.symmetry_group_size(d) for d, c in zip(dim, self.cells))

    def cell_orientation_reflection_map(self):
        """Return the map indicating whether each possible cell orientation causes reflection (``1``) or not (``0``)."""
        return make_cell_orientation_reflection_map_tensorproduct(self.cells)

    def compare(self, op, other):
        """Parent-based comparison between simplicial complexes.
        This is done dimension by dimension."""
        if hasattr(other, "product"):
            other = other.product
        if isinstance(other, type(self)):
            return all(op(a, b) for a, b in zip(self.cells, other.cells))
        else:
            return op(self, other)

    def __gt__(self, other):
        return self.compare(operator.gt, other)

    def __lt__(self, other):
        return self.compare(operator.lt, other)

    def __ge__(self, other):
        return self.compare(operator.ge, other)

    def __le__(self, other):
        return self.compare(operator.le, other)

    def extract_extrinsic_orientation(self, o):
        """Extract extrinsic orientation.

        Parameters
        ----------
        o : Orientation
            Total orientation.

        Returns
        -------
        Orientation
            Extrinsic orientation.

        Notes
        -----
        The difinition of orientations used here must be consistent with
        that used in make_entity_permutations_tensorproduct.

        """
        if not isinstance(o, Orientation):
            raise TypeError(f"Expecting an instance of Orientation : got {o}")
        dim = len(self.cells)
        size_io = 2  # Number of possible intrinsic orientations along each axis.
        return o // size_io**dim

    def extract_intrinsic_orientation(self, o, axis):
        """Extract intrinsic orientation.

        Parameters
        ----------
        o : Orientation
            Total orientation. ``//`` and ``%`` must be overloaded in type(o).
        axis : int
            Reference cell axis for which intrinsic orientation is computed.

        Returns
        -------
        Orientation
            Intrinsic orientation.

        Notes
        -----
        Must be consistent with make_entity_permutations_tensorproduct.

        """
        if not isinstance(o, Orientation):
            raise TypeError(f"Expecting an instance of Orientation : got {o}")
        dim = len(self.cells)
        if axis >= dim:
            raise ValueError(f"Must give 0 <= axis < {dim} : got {axis}")
        size_io = 2  # Number of possible intrinsic orientations along each axis.
        return o % size_io**dim // size_io**(dim - 1 - axis) % size_io

    @property
    def extrinsic_orientation_permutation_map(self):
        """A map from extrinsic orientations to corresponding axis permutation matrices.

        Notes
        -----
        result[eo] gives the physical axis-reference axis permutation matrix corresponding to
        eo (extrinsic orientation).

        """
        dim = len(self.cells)
        a = numpy.zeros((factorial(dim), dim, dim), dtype=int)
        ai = numpy.array(list(make_entity_permutations_simplex(dim - 1, 2).values()), dtype=int).reshape((factorial(dim), dim, 1))
        numpy.put_along_axis(a, ai, 1, axis=2)
        return a


class Hypercube(Cell):
    """Abstract class for a reference hypercube"""

    def __init__(self, dimension, product):
        self.dimension = dimension
        self.shape = hypercube_shapes[dimension]

        pt = product.get_topology()
        verts = product.get_vertices()
        topology = flatten_entities(pt)

        super().__init__(self.shape, verts, topology)

        self.product = product
        self.unflattening_map = compute_unflattening_map(pt)

    def get_dimension(self):
        """Returns the subelement dimension of the cell.  Same as the
        spatial dimension."""
        return self.get_spatial_dimension()

    def construct_subelement(self, dimension):
        """Constructs the reference element of a cell subentity
        specified by subelement dimension.

        :arg dimension: subentity dimension (integer)
        """
        sd = self.get_spatial_dimension()
        if dimension > sd:
            raise ValueError(f"Invalid dimension: {(dimension,)}")
        elif dimension == sd:
            return self
        else:
            sub_element = self.product.construct_subelement((dimension,) + (0,)*(len(self.product.cells) - 1))
            return flatten_reference_cube(sub_element)

    def get_entity_transform(self, dim, entity_i):
        """Returns a mapping of point coordinates from the
        `entity_i`-th subentity of dimension `dim` to the cell.

        :arg dim: entity dimension (integer)
        :arg entity_i: entity number (integer)
        """
        d, e = self.unflattening_map[(dim, entity_i)]
        return self.product.get_entity_transform(d, e)

    def volume(self):
        """Computes the volume in the appropriate dimensional measure."""
        return self.product.volume()

    def compute_reference_normal(self, facet_dim, facet_i):
        """Returns the unit normal in infinity norm to facet_i."""
        sd = self.get_spatial_dimension()
        assert facet_dim == sd - 1
        d, i = self.unflattening_map[(facet_dim, facet_i)]
        return self.product.compute_reference_normal(d, i)

    def contains_point(self, point, epsilon=0):
        """Checks if reference cell contains given point
        (with numerical tolerance as given by the L1 distance (aka 'manhattan',
        'taxicab' or rectilinear distance) to the cell.

        Parameters
        ----------
        point : numpy.ndarray, list or symbolic expression
            The coordinates of the point.
        epsilon : float
            The tolerance for the check.

        Returns
        -------
        bool : True if the point is inside the cell, False otherwise.

        """
        return self.product.contains_point(point, epsilon=epsilon)

    def distance_to_point_l1(self, point, rescale=False):
        """Get the L1 distance (aka 'manhattan', 'taxicab' or rectilinear
        distance) to a point with 0.0 if the point is inside the cell.

        For more information see the docstring for the UFCSimplex method."""
        return self.product.distance_to_point_l1(point, rescale=rescale)

    def symmetry_group_size(self, dim):
        """Size of hypercube symmetry group is d! * 2**d"""
        return factorial(dim) * (2**dim)

    def cell_orientation_reflection_map(self):
        """Return the map indicating whether each possible cell orientation causes reflection (``1``) or not (``0``)."""
        return self.product.cell_orientation_reflection_map()

    def __gt__(self, other):
        return self.product > other

    def __lt__(self, other):
        return self.product < other

    def __ge__(self, other):
        return self.product >= other

    def __le__(self, other):
        return self.product <= other


class UFCHypercube(Hypercube):
    """Reference UFC Hypercube

    UFCHypercube: [0, 1]^d with vertices in
    lexicographical order."""

    def __init__(self, dim):
        cells = [UFCInterval()] * dim
        product = TensorProductCell(*cells)
        super().__init__(dim, product)

    def construct_subelement(self, dimension):
        """Constructs the reference element of a cell subentity
        specified by subelement dimension.

        :arg dimension: subentity dimension (integer)
        """
        sd = self.get_spatial_dimension()
        if dimension > sd:
            raise ValueError(f"Invalid dimension: {dimension}")
        elif dimension == sd:
            return self
        else:
            return ufc_hypercube(dimension)


class UFCQuadrilateral(UFCHypercube):
    r"""This is the reference quadrilateral with vertices
    (0.0, 0.0), (0.0, 1.0), (1.0, 0.0) and (1.0, 1.0).

    Orientation of a physical cell is computed systematically
    by comparing the canonical orderings of its facets and
    the facets in the FIAT reference cell.

    As an example, we compute the orientation of a
    quadrilateral cell:

       +---3---+           +--57---+
       |       |           |       |
       0       1          43       55
       |       |           |       |
       +---2---+           +--42---+
    FIAT canonical     Mapped example physical cell

    Suppose that the facets of the physical cell
    are canonically ordered as:

    C = [55, 42, 43, 57]

    FIAT index to Physical index map must be such that
    C[0] = 55 is mapped to a vertical facet; in this
    example it is:

    M = [43, 55, 42, 57]

    C and M are decomposed into "vertical" and "horizontal"
    parts, keeping the relative orders of numbers:

    C -> C0 = [55, 43], C1 = [42, 57]
    M -> M0 = [43, 55], M1 = [42, 57]

    Then the orientation of the cell is computed as the
    following:

    C0.index(M0[0]) = 1; C0.remove(M0[0])
    C0.index(M0[1]) = 0; C0.remove(M0[1])
    C1.index(M1[0]) = 0; C1.remove(M1[0])
    C1.index(M1[1]) = 0; C1.remove(M1[1])

    o = 2 * 1 + 0 = 2
    """

    def __init__(self):
        super().__init__(2)


class UFCHexahedron(UFCHypercube):
    """This is the reference hexahedron with vertices
    (0.0, 0.0, 0.0), (0.0, 0.0, 1.0), (0.0, 1.0, 0.0), (0.0, 1.0, 1.0),
    (1.0, 0.0, 0.0), (1.0, 0.0, 1.0), (1.0, 1.0, 0.0) and (1.0, 1.0, 1.0)."""

    def __init__(self):
        super().__init__(3)


def make_affine_mapping(xs, ys):
    """Constructs (A,b) such that x --> A * x + b is the affine
    mapping from the simplex defined by xs to the simplex defined by ys."""

    dim_x = len(xs[0])
    dim_y = len(ys[0])

    if len(xs) != len(ys):
        raise Exception("")

    # find A in R^{dim_y,dim_x}, b in R^{dim_y} such that
    # A xs[i] + b = ys[i] for all i

    mat = numpy.zeros((dim_x * dim_y + dim_y, dim_x * dim_y + dim_y), "d")
    rhs = numpy.zeros((dim_x * dim_y + dim_y,), "d")

    # loop over points
    for i in range(len(xs)):
        # loop over components of each A * point + b
        for j in range(dim_y):
            row_cur = i * dim_y + j
            col_start = dim_x * j
            col_finish = col_start + dim_x
            mat[row_cur, col_start:col_finish] = numpy.array(xs[i])
            rhs[row_cur] = ys[i][j]
            # need to get terms related to b
            mat[row_cur, dim_y * dim_x + j] = 1.0

    sol = numpy.linalg.solve(mat, rhs)

    A = numpy.reshape(sol[:dim_x * dim_y], (dim_y, dim_x))
    b = sol[dim_x * dim_y:]

    return A, b


def ufc_hypercube(spatial_dim):
    """Factory function that maps spatial dimension to an instance of
    the UFC reference hypercube of that dimension."""
    if spatial_dim == 0:
        return Point()
    elif spatial_dim == 1:
        return UFCInterval()
    elif spatial_dim == 2:
        return UFCQuadrilateral()
    elif spatial_dim == 3:
        return UFCHexahedron()
    else:
        raise RuntimeError(f"Can't create UFC hypercube of dimension {spatial_dim}.")


def default_simplex(spatial_dim):
    """Factory function that maps spatial dimension to an instance of
    the default reference simplex of that dimension."""
    if spatial_dim == 0:
        return Point()
    elif spatial_dim == 1:
        return DefaultLine()
    elif spatial_dim == 2:
        return DefaultTriangle()
    elif spatial_dim == 3:
        return DefaultTetrahedron()
    else:
        raise RuntimeError(f"Can't create default simplex of dimension {spatial_dim}.")


def ufc_simplex(spatial_dim):
    """Factory function that maps spatial dimension to an instance of
    the UFC reference simplex of that dimension."""
    if spatial_dim == 0:
        return Point()
    elif spatial_dim == 1:
        return UFCInterval()
    elif spatial_dim == 2:
        return UFCTriangle()
    elif spatial_dim == 3:
        return UFCTetrahedron()
    else:
        raise RuntimeError(f"Can't create UFC simplex of dimension {spatial_dim}.")


def symmetric_simplex(spatial_dim):
    A = numpy.array([[2, 1, 1],
                     [0, numpy.sqrt(3), numpy.sqrt(3)/3],
                     [0, 0, numpy.sqrt(6)*(2/3)]])
    A = A[:spatial_dim, :][:, :spatial_dim]
    b = A.sum(axis=1) * (-1 / (1 + spatial_dim))
    Ref1 = ufc_simplex(spatial_dim)
    v = numpy.dot(Ref1.get_vertices(), A.T) + b[None, :]
    vertices = tuple(map(tuple, v))
    return SymmetricSimplex(Ref1.get_shape(), vertices, Ref1.get_topology())


def ufc_cell(cell):
    """Handle incoming calls from FFC."""

    # celltype could be a string or a cell.
    if isinstance(cell, str):
        celltype = cell
    else:
        celltype = cell.cellname()

    if " * " in celltype:
        # Tensor product cell
        return TensorProductCell(*map(ufc_cell, celltype.split(" * ")))
    elif celltype == "quadrilateral":
        return UFCQuadrilateral()
    elif celltype == "hexahedron":
        return UFCHexahedron()
    elif celltype == "vertex":
        return ufc_simplex(0)
    elif celltype == "interval":
        return ufc_simplex(1)
    elif celltype == "triangle":
        return ufc_simplex(2)
    elif celltype == "tetrahedron":
        return ufc_simplex(3)
    else:
        raise RuntimeError(f"Don't know how to create UFC cell of type {str(celltype)}")


def volume(verts):
    """Constructs the volume of the simplex spanned by verts"""

    # use fact that volume of UFC reference element is 1/n!
    sd = len(verts) - 1
    ufcel = ufc_simplex(sd)
    ufcverts = ufcel.get_vertices()

    A, b = make_affine_mapping(ufcverts, verts)

    # can't just take determinant since, e.g. the face of
    # a tet being mapped to a 2d triangle doesn't have a
    # square matrix

    (u, s, vt) = numpy.linalg.svd(A)

    # this is the determinant of the "square part" of the matrix
    # (ie the part that maps the restriction of the higher-dimensional
    # stuff to UFC element
    p = numpy.prod([si for si in s if (si) > 1.e-10])

    return p / factorial(sd)


def tuple_sum(tree):
    """
    This function calculates the sum of elements in a tuple, it is needed to handle nested tuples in TensorProductCell.
    Example: tuple_sum(((1, 0), 1)) returns 2
    If input argument is not the tuple, returns input.
    """
    if isinstance(tree, tuple):
        return sum(map(tuple_sum, tree))
    else:
        return tree


def is_ufc(cell):
    if isinstance(cell, (Point, UFCInterval, UFCHypercube, UFCSimplex)):
        return True
    elif isinstance(cell, TensorProductCell):
        return all(is_ufc(c) for c in cell.cells)
    else:
        return False


def is_hypercube(cell):
    if isinstance(cell, (DefaultLine, UFCInterval, Hypercube)):
        return True
    elif isinstance(cell, TensorProductCell):
        return all(is_hypercube(c) for c in cell.cells)
    else:
        return False


def flatten_reference_cube(ref_el):
    """This function flattens a Tensor Product hypercube to the corresponding UFC hypercube"""
    if ref_el.get_spatial_dimension() <= 1:
        # Just return point/interval cell arguments
        return ref_el
    else:
        # Handle cases where cell is a quad/cube constructed from a tensor product or
        # an already flattened element
        if isinstance(ref_el, TensorProductCell):
            if is_ufc(ref_el):
                return ufc_hypercube(ref_el.get_spatial_dimension())
            return Hypercube(ref_el.get_spatial_dimension(), ref_el)
        elif is_hypercube(ref_el):
            return ref_el
        else:
            raise TypeError('Invalid cell type')


def flatten_entities(topology_dict):
    """This function flattens topology dict of TensorProductCell and entity_dofs dict of TensorProductElement"""

    flattened_entities = defaultdict(list)
    for dim in sorted(topology_dict.keys()):
        flat_dim = tuple_sum(dim)
        flattened_entities[flat_dim] += [v for k, v in sorted(topology_dict[dim].items())]

    return {dim: dict(enumerate(entities))
            for dim, entities in flattened_entities.items()}


def flatten_permutations(perm_dict):
    """This function flattens permutation dict of TensorProductElement"""

    flattened_permutations = defaultdict(list)
    for dim in sorted(perm_dict.keys()):
        flat_dim = tuple_sum(dim)
        flattened_permutations[flat_dim] += [{o: v[o_tuple] for o, o_tuple in enumerate(sorted(v))}
                                             for k, v in sorted(perm_dict[dim].items())]
    return {dim: dict(enumerate(perms))
            for dim, perms in flattened_permutations.items()}


def compute_unflattening_map(topology_dict):
    """This function returns unflattening map for the given tensor product topology dict."""

    counter = defaultdict(count)
    unflattening_map = {}

    for dim, entities in sorted(topology_dict.items()):
        flat_dim = tuple_sum(dim)
        for entity in entities:
            flat_entity = next(counter[flat_dim])
            unflattening_map[(flat_dim, flat_entity)] = (dim, entity)

    return unflattening_map


def max_complex(complexes):
    max_cell = max(complexes)
    if all(max_cell >= b for b in complexes):
        return max_cell
    else:
        raise ValueError("Cannot find the maximal complex")



# Copyright (C) 2015-2016 Jan Blechta, Andrew T T McRae, and others
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from FIAT.dual_set import DualSet
from FIAT.finite_element import CiarletElement


class RestrictedDualSet(DualSet):
    """Restrict the given DualSet to the specified list of dofs."""

    def __init__(self, dual, indices):
        indices = list(sorted(indices))
        ref_el = dual.get_reference_element()
        nodes_old = dual.get_nodes()
        entity_ids = {}
        nodes = []
        for d, entities in dual.get_entity_ids().items():
            entity_ids[d] = {}
            for entity, dofs in entities.items():
                entity_ids[d][entity] = [indices.index(dof)
                                         for dof in dofs if dof in indices]
        nodes = [nodes_old[i] for i in indices]
        self._dual = dual
        super().__init__(nodes, ref_el, entity_ids)

    def get_indices(self, restriction_domain, take_closure=True):
        """Return the list of dofs with support on a given restriction domain.

        :arg restriction_domain: can be 'interior', 'vertex', 'edge', 'face' or 'facet'
        :kwarg take_closure: Are we taking the closure of the restriction domain?
        """
        # Call get_indices on the parent class to support multiple restriction domains
        return type(self._dual).get_indices(self, restriction_domain, take_closure=take_closure)


class RestrictedElement(CiarletElement):
    """Restrict the given element to the specified list of dofs."""

    def __init__(self, element, indices=None, restriction_domain=None, take_closure=True):
        '''For sake of argument, indices overrides restriction_domain'''

        if not (indices or restriction_domain):
            raise RuntimeError("Either indices or restriction_domain must be passed in")

        if not indices:
            indices = element.dual.get_indices(restriction_domain, take_closure=take_closure)

        if isinstance(indices, str):
            raise RuntimeError("variable 'indices' was a string; did you forget to use a keyword?")

        if len(indices) == 0:
            raise ValueError("No point in creating empty RestrictedElement.")

        self._element = element
        self._indices = indices

        # Restrict primal set
        poly_set = element.get_nodal_basis().take(indices)

        # Restrict dual set
        dual = RestrictedDualSet(element.get_dual_set(), indices)

        # Restrict mapping
        mapping_old = element.mapping()
        mapping_new = [mapping_old[dof] for dof in indices]
        assert all(e_mapping == mapping_new[0] for e_mapping in mapping_new)

        # Call constructor of CiarletElement
        super().__init__(poly_set, dual, 0, element.get_formdegree(), mapping_new[0])



# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Written by Pablo D. Brubeck (brubeck@protonmail.com), 2024

# This is not quite Bernardi-Raugel, but it has 2*dim*(dim+1) dofs and includes
# dim**2-1 extra constraint functionals.  The first (dim+1)**2 basis functions
# are the reference element bfs, but the extra dim**2-1 are used in the
# transformation theory.

from FIAT import finite_element, dual_set, polynomial_set, expansions
from FIAT.functional import ComponentPointEvaluation, FrobeniusIntegralMoment
from FIAT.hierarchical import make_dual_bubbles
from FIAT.quadrature import FacetQuadratureRule

import numpy
import math


def BernardiRaugelSpace(ref_el, order):
    """Return a basis for the extended Bernardi-Raugel space: (Pk + FacetBubble)^d."""
    sd = ref_el.get_spatial_dimension()
    if order > sd:
        raise ValueError("The Bernardi-Raugel space is only defined for order <= dim")
    Pd = polynomial_set.ONPolynomialSet(ref_el, sd, shape=(sd,), scale=1, variant="bubble")
    dimPd = expansions.polynomial_dimension(ref_el, sd, continuity="C0")
    entity_ids = expansions.polynomial_entity_ids(ref_el, sd, continuity="C0")

    slices = {dim: slice(math.comb(order-1, dim)) for dim in range(order)}
    slices.pop(sd-1, None)
    ids = [i + j * dimPd
           for dim in slices
           for f in sorted(entity_ids[dim])
           for i in entity_ids[dim][f][slices[dim]]
           for j in range(sd)]

    interior_facets = ref_el.get_interior_facets(sd-1) or ()
    facets = list(set(entity_ids[sd-1]) - set(interior_facets))
    ids.extend(i + j*dimPd
               for f in sorted(facets)
               for i in entity_ids[sd-1][f]
               for j in range(sd))
    return Pd.take(ids)


class BernardiRaugelDualSet(dual_set.DualSet):
    """The Bernardi-Raugel dual set."""
    def __init__(self, ref_el, order=1, degree=None, reduced=False, ref_complex=None):
        if ref_complex is None:
            ref_complex = ref_el
        sd = ref_el.get_spatial_dimension()
        if degree is None:
            degree = sd
        if order > sd:
            raise ValueError(f"{type(self).__name__} is only defined for order <= dim")
        top = ref_el.get_topology()
        entity_ids = {dim: {entity: [] for entity in sorted(top[dim])} for dim in sorted(top)}

        nodes = []
        if order > 0:
            # Point evaluation at lattice points
            for dim in sorted(top):
                for entity in sorted(top[dim]):
                    cur = len(nodes)
                    pts = ref_el.make_points(dim, entity, order)
                    nodes.extend(ComponentPointEvaluation(ref_el, comp, (sd,), pt)
                                 for pt in pts for comp in range(sd))
                    entity_ids[dim][entity].extend(range(cur, len(nodes)))

        if order < sd:
            # Face moments of normal/tangential components against dual bubbles
            ref_facet = ref_complex.construct_subcomplex(sd-1)
            codim = sd-1 if degree == 1 and ref_facet.is_macrocell() else 0
            Q, phis = make_dual_bubbles(ref_facet, degree, codim=codim)
            f_at_qpts = phis[-1]
            if codim != 0:
                f_at_qpts -= numpy.dot(f_at_qpts, Q.get_weights()) / ref_facet.volume()

            interior_facets = ref_el.get_interior_facets(sd-1) or ()
            facets = list(set(top[sd-1]) - set(interior_facets))

            Qs = {f: FacetQuadratureRule(ref_el, sd-1, f, Q) for f in facets}
            thats = {f: ref_el.compute_tangents(sd-1, f) for f in facets}

            R = numpy.array([[0, 1], [-1, 0]])
            ndir = 1 if reduced else sd
            for i in range(ndir):
                for f in sorted(facets):
                    cur = len(nodes)
                    if i == 0:
                        udir = numpy.dot(R, *thats[f]) if sd == 2 else numpy.cross(*thats[f])
                    else:
                        udir = thats[f][i-1]
                    detJ = Qs[f].jacobian_determinant()
                    phi_at_qpts = udir[:, None] * f_at_qpts[None, :] / detJ
                    nodes.append(FrobeniusIntegralMoment(ref_el, Qs[f], phi_at_qpts))
                    entity_ids[sd-1][f].extend(range(cur, len(nodes)))
        super().__init__(nodes, ref_el, entity_ids)


class BernardiRaugel(finite_element.CiarletElement):
    """The Bernardi-Raugel (extended) element.

    This element does not belong to a Stokes complex, but can be paired with
    DG_{k-1}. This pair is inf-sup stable, but only weakly divergence-free.
    """
    def __init__(self, ref_el, order=1):
        degree = ref_el.get_spatial_dimension()
        if order >= degree:
            raise ValueError(f"{type(self).__name__} only defined for order < dim")
        poly_set = BernardiRaugelSpace(ref_el, order)
        dual = BernardiRaugelDualSet(ref_el, order, degree=degree)
        formdegree = 0
        super().__init__(poly_set, dual, degree, formdegree, mapping="contravariant piola")



import pytest
import numpy

from FIAT import (HsiehCloughTocher as HCT,
                  AlfeldSorokina as AS,
                  ArnoldQin as AQ,
                  Lagrange as CG,
                  DiscontinuousLagrange as DG)
from FIAT.reference_element import ufc_simplex

from FIAT.polynomial_set import ONPolynomialSet
from FIAT.macro import CkPolynomialSet
from FIAT.alfeld_sorokina import AlfeldSorokinaSpace
from FIAT.arnold_qin import ArnoldQinSpace
from FIAT.christiansen_hu import ChristiansenHuSpace
from FIAT.guzman_neilan import (
    GuzmanNeilanFirstKindH1,
    GuzmanNeilanSecondKindH1,
    GuzmanNeilanH1div,
    GuzmanNeilanSpace)
from FIAT.restricted import RestrictedElement


T = ufc_simplex(2)
S = ufc_simplex(3)


def rHCT(cell):
    return RestrictedElement(HCT(cell, reduced=True), restriction_domain="vertex")


def rAQ(cell):
    return RestrictedElement(AQ(cell, reduced=True), indices=list(range(9)))


def span_greater_equal(A, B):
    # span(A) >= span(B)
    _, residual, *_ = numpy.linalg.lstsq(A.reshape(A.shape[0], -1).T,
                                         B.reshape(B.shape[0], -1).T)
    return numpy.allclose(residual, 0)


def span_equal(A, B):
    # span(A) == span(B)
    return span_greater_equal(A, B) and span_greater_equal(B, A)


def div(U):
    """Return divergence from dict of tabulations """
    return sum(U[k][:, k.index(1), :] for k in U if sum(k) == 1)


def rot(U):
    """Return rot from dict of tabulations """
    return numpy.stack([U[(0, 1)], -U[(1, 0)]], axis=1)


def make_points(K, degree):
    top = K.get_topology()
    pts = []
    for dim in top:
        for entity in top[dim]:
            pts.extend(K.make_points(dim, entity, degree))
    return pts


def check_h1div_space(V, degree, reduced=False, bubble=False):
    # Test that the divergence of the polynomial space V is spanned by a C0 basis
    A = V.get_reference_element()
    sd = A.get_spatial_dimension()
    z = (0,)*sd

    pts = make_points(A, degree+2)
    V_tab = V.tabulate(pts, 1)
    V_div = div(V_tab)

    C0 = CkPolynomialSet(A, degree-1, order=0, variant="bubble")
    C0_tab = C0.tabulate(pts)[z]
    assert span_equal(V_div, C0_tab)

    if bubble:
        # Test that div(Bubbles) = C0 int H^1_0
        assert span_equal(V_div[-(sd+1):], C0_tab[-1:])

    k = degree - 1 if reduced else degree
    # Test that V includes Pk
    cell = A.get_parent() or A
    Pk = ONPolynomialSet(cell, k, shape=(sd,))
    Pk_tab = Pk.tabulate(pts)[z]
    assert span_greater_equal(V_tab[z], Pk_tab)


@pytest.mark.parametrize("cell", (T, S))
@pytest.mark.parametrize("degree", (2, 3, 4))
def test_h1div_alfeld_sorokina(cell, degree):
    # Test that the divergence of the Alfeld-Sorokina space is spanned by a C0 basis
    V = AlfeldSorokinaSpace(cell, degree)
    check_h1div_space(V, degree)


@pytest.mark.parametrize("cell", (S,))
@pytest.mark.parametrize("reduced", (False, True), ids=("full", "reduced"))
def test_h1div_guzman_neilan(cell, reduced):
    # Test that the divergence of AS + GN Bubble is spanned by a C0 basis
    sd = cell.get_spatial_dimension()
    degree = 2
    fe = GuzmanNeilanH1div(cell, degree, reduced=reduced)
    reduced_dim = fe.space_dimension() - (sd-1)*(sd+1)
    V = fe.get_nodal_basis().take(list(range(reduced_dim)))
    check_h1div_space(V, degree, reduced=reduced, bubble=True)


def check_stokes_complex(spaces, degree):
    # Test that we have a discrete Stokes complex, verifying that the range of
    # the exterior derivative of each space is contained by the next space in
    # the sequence
    A = spaces[0].get_reference_complex()
    sd = A.get_spatial_dimension()
    z = (0,) * sd

    pts = make_points(A, degree+2)
    tab = [V.tabulate(1, pts) for V in spaces]
    if len(tab) > 2:
        # check rot(V0) in V1
        assert span_greater_equal(tab[1][z], rot(tab[0]))

    # check div(V1) = V2
    assert span_equal(tab[-1][z], div(tab[-2]))

    # Test that V1 includes Pk
    cell = A.get_parent() or A
    Pk = ONPolynomialSet(cell, degree, shape=(sd,))
    Pk_tab = Pk.tabulate(pts)[z]
    assert span_greater_equal(tab[-2][z], Pk_tab)


@pytest.mark.parametrize("reduced", (False, True), ids=("full", "reduced"))
@pytest.mark.parametrize("sobolev", ("H1", "H1div"))
@pytest.mark.parametrize("cell", (T,))
def test_hct_stokes_complex(cell, sobolev, reduced):
    if sobolev == "H1":
        if reduced:
            spaces = [rHCT(cell), rAQ(cell), DG(cell, 0)]
            degree = 1
        else:
            spaces = [HCT(cell), AQ(cell), DG(cell, 0)]
            degree = 1
    elif sobolev == "H1div":
        if reduced:
            spaces = [rHCT(cell),
                      GuzmanNeilanH1div(cell, reduced=True),
                      CG(cell, 1, variant="alfeld")]
            degree = 1
        else:
            spaces = [HCT(cell), AS(cell), CG(cell, 1, variant="alfeld")]
            degree = 2
    else:
        raise ValueError(f"Unexpected sobolev space {sobolev}")
    check_stokes_complex(spaces, degree)


@pytest.mark.parametrize("cell", (T, S))
@pytest.mark.parametrize("kind", (1, 2, "H1div", "H1div-red"))
def test_gn_stokes_pairs(cell, kind):
    order = cell.get_spatial_dimension() - 1
    if kind == 1:
        spaces = [GuzmanNeilanFirstKindH1(cell, order), DG(cell, order-1)]
        degree = order
    elif kind == 2:
        spaces = [GuzmanNeilanSecondKindH1(cell, order), DG(cell, order-1, variant="alfeld")]
        degree = order
    elif kind == "H1div":
        spaces = [GuzmanNeilanH1div(cell), CG(cell, 1, variant="alfeld")]
        degree = 2
    elif kind == "H1div-red":
        spaces = [GuzmanNeilanH1div(cell, reduced=True), CG(cell, 1, variant="alfeld")]
        degree = 1
    else:
        raise ValueError(f"Unexpected kind {kind}")
    check_stokes_complex(spaces, degree)


@pytest.mark.parametrize("cell", (T, S))
@pytest.mark.parametrize("family", ("AQ", "CH", "GN", "GN2"))
def test_minimal_stokes_space(cell, family):
    # Test that the C0 Stokes space is spanned by a C0 basis
    # Also test that its divergence is constant
    sd = cell.get_spatial_dimension()
    if family == "GN":
        degree = 1
        space = GuzmanNeilanSpace
    elif family == "GN2":
        degree = 1
        space = lambda *args, **kwargs: GuzmanNeilanSpace(*args, kind=2, **kwargs)
    elif family == "CH":
        degree = 1
        space = ChristiansenHuSpace
    elif family == "AQ":
        if sd != 2:
            return
        degree = 2
        space = ArnoldQinSpace

    W = space(cell, degree)
    V = space(cell, degree, reduced=True)
    Wdim = W.get_num_members()
    Vdim = V.get_num_members()
    K = W.get_reference_element()
    sd = K.get_spatial_dimension()

    pts = make_points(K, degree+2)
    C0 = CkPolynomialSet(K, sd, order=0, variant="bubble")
    C0_tab = C0.tabulate(pts)
    Wtab = W.tabulate(pts, 1)
    Vtab = V.tabulate(pts, 1)
    z = (0,) * sd
    for Xtab in (Vtab, Wtab):
        # Test that the space is full rank
        _, sig, _ = numpy.linalg.svd(Xtab[z].reshape(-1, sd*len(pts)).T, full_matrices=True)
        assert all(sig > 1E-10)

        # Test that the space is C0
        for k in range(sd):
            _, residual, *_ = numpy.linalg.lstsq(C0_tab[z].T, Xtab[z][:, k, :].T)
            assert numpy.allclose(residual, 0)

        # Test that divergence is in P0
        divX = div(Xtab)[:Vdim]
        if family in {"GN", "GN2"}:
            # Test that div(GN2) is in P0(Alfeld)
            ref_el = K if family == "GN2" else K.get_parent()
            P0 = ONPolynomialSet(ref_el, degree-1)
            P0_tab = P0.tabulate(pts)[z]
            assert span_equal(divX, P0_tab)
        else:
            assert numpy.allclose(divX, divX[:, 0][:, None])

    # Test that the full space includes the reduced space
    assert Wdim > Vdim
    assert span_greater_equal(Wtab[z], Vtab[z])



"""Utilities for building an Impero AST from an ordered list of
terminal Impero operations, and for building any additional data
required for straightforward C code generation.

What this module does is independent of the generated code target.
"""

import collections
from functools import singledispatch
from itertools import chain, groupby

from gem.node import traversal, collect_refcount
from gem import gem, impero as imp, optimise, scheduling


# ImperoC is named tuple for C code generation.
#
# Attributes:
#     tree        - Impero AST describing the loop structure and operations
#     temporaries - List of GEM expressions which have assigned temporaries
#     declare     - Where to declare temporaries to get correct C code
#     indices     - Indices for declarations and referencing values
ImperoC = collections.namedtuple('ImperoC', ['tree', 'temporaries', 'declare', 'indices'])


class NoopError(Exception):
    """No operations in the kernel."""
    pass


def preprocess_gem(expressions, replace_delta=True, remove_componenttensors=True):
    """Lower GEM nodes that cannot be translated to C directly."""
    if remove_componenttensors:
        expressions = optimise.remove_componenttensors(expressions)
    if replace_delta:
        expressions = optimise.replace_delta(expressions)
    return expressions


def compile_gem(assignments, prefix_ordering, remove_zeros=False,
                emit_return_accumulate=True):
    """Compiles GEM to Impero.

    :arg assignments: list of (return variable, expression DAG root) pairs
    :arg prefix_ordering: outermost loop indices
    :arg remove_zeros: remove zero assignment to return variables
    :arg emit_return_accumulate: emit ReturnAccumulate nodes (see
         :func:`~.scheduling.emit_operations`)? If False,
         split into Accumulate/Return pairs. Set to False if the
         output tensor of kernels is not guaranteed to be zero on entry.
    """
    # Remove zeros
    if remove_zeros:
        def nonzero(assignment):
            variable, expression = assignment
            return not isinstance(expression, gem.Zero)
        assignments = list(filter(nonzero, assignments))

    # Just the expressions
    expressions = [expression for variable, expression in assignments]

    # Collect indices in a deterministic order
    indices = list(collections.OrderedDict.fromkeys(chain.from_iterable(
        node.index_ordering()
        for node in traversal(expressions)
        if isinstance(node, (gem.Indexed, gem.FlexiblyIndexed))
    )))

    # Build ordered index map
    index_ordering = make_prefix_ordering(indices, prefix_ordering)
    apply_ordering = make_index_orderer(index_ordering)

    get_indices = lambda expr: apply_ordering(expr.free_indices)

    # Build operation ordering
    ops = scheduling.emit_operations(assignments, get_indices, emit_return_accumulate)

    # Empty kernel
    if len(ops) == 0:
        raise NoopError()

    # Drop unnecessary temporaries
    ops = inline_temporaries(expressions, ops)

    # Build Impero AST
    tree = make_loop_tree(ops, get_indices)

    # Collect temporaries
    temporaries = collect_temporaries(tree)

    # Determine declarations
    declare, indices = place_declarations(tree, temporaries, get_indices)

    # Prepare ImperoC (Impero AST + other data for code generation)
    return ImperoC(tree, temporaries, declare, indices)


def make_prefix_ordering(indices, prefix_ordering):
    """Creates an ordering of ``indices`` which starts with those
    indices in ``prefix_ordering``."""
    # Need to return deterministically ordered indices
    return tuple(prefix_ordering) + tuple(k for k in indices if k not in prefix_ordering)


def make_index_orderer(index_ordering):
    """Returns a function which given a set of indices returns those
    indices in the order as they appear in ``index_ordering``."""
    idx2pos = {idx: pos for pos, idx in enumerate(index_ordering)}

    def apply_ordering(indices):
        return tuple(sorted(indices, key=lambda i: idx2pos[i]))
    return apply_ordering


def inline_temporaries(expressions, ops):
    """Inline temporaries which could be inlined without blowing up
    the code.

    :arg expressions: a multi-root GEM expression DAG, used for
                      reference counting
    :arg ops: ordered list of Impero terminals
    :returns: a filtered ``ops``, without the unnecessary
              :class:`impero.Evaluate`s
    """
    refcount = collect_refcount(expressions)

    candidates = set()  # candidates for inlining
    for op in ops:
        if isinstance(op, imp.Evaluate):
            expr = op.expression
            if expr.shape == () and refcount[expr] == 1:
                candidates.add(expr)

    # Prevent inlining that pulls expressions into inner loops
    for node in traversal(expressions):
        for child in node.children:
            if child in candidates and set(child.free_indices) < set(node.free_indices):
                candidates.remove(child)

    # Filter out candidates
    return [op for op in ops if not (isinstance(op, imp.Evaluate) and op.expression in candidates)]


def collect_temporaries(tree):
    """Collects GEM expressions to assign to temporaries from a list
    of Impero terminals."""
    result = []
    for node in traversal((tree,)):
        # IndexSum temporaries should be added either at Initialise or
        # at Accumulate.  The difference is only in ordering
        # (numbering).  We chose Accumulate here.
        if isinstance(node, imp.Accumulate):
            result.append(node.indexsum)
        elif isinstance(node, imp.Evaluate):
            result.append(node.expression)
    return result


def make_loop_tree(ops, get_indices, level=0):
    """Creates an Impero AST with loops from a list of operations and
    their respective free indices.

    :arg ops: a list of Impero terminal nodes
    :arg get_indices: callable mapping from GEM nodes to an ordering
                      of free indices
    :arg level: depth of loop nesting
    :returns: Impero AST with loops, without declarations
    """
    keyfunc = lambda op: op.loop_shape(get_indices)[level:level+1]
    statements = []
    for first_index, op_group in groupby(ops, keyfunc):
        if first_index:
            inner_block = make_loop_tree(op_group, get_indices, level+1)
            statements.append(imp.For(first_index[0], inner_block))
        else:
            statements.extend(op_group)
    # Remove no-op terminals from the tree
    statements = [s for s in statements if not isinstance(s, imp.Noop)]
    return imp.Block(statements)


def place_declarations(tree, temporaries, get_indices):
    """Determines where and how to declare temporaries for an Impero AST.

    :arg tree: Impero AST to determine the declarations for
    :arg temporaries: list of GEM expressions which are assigned to
                      temporaries
    :arg get_indices: callable mapping from GEM nodes to an ordering
                      of free indices
    """
    numbering = {t: n for n, t in enumerate(temporaries)}
    assert len(numbering) == len(temporaries)

    # Collect the total number of temporary references
    total_refcount = collections.Counter()
    for node in traversal((tree,)):
        if isinstance(node, imp.Terminal):
            total_refcount.update(temp_refcount(numbering, node))
    assert set(total_refcount) == set(temporaries)

    # Result
    declare = {}
    indices = {}

    @singledispatch
    def recurse(expr, loop_indices):
        """Visit an Impero AST to collect declarations.

        :arg expr: Impero tree node
        :arg loop_indices: loop indices (in order) from the outer
                           loops surrounding ``expr``
        :returns: :class:`collections.Counter` with the reference
                  counts for each temporary in the subtree whose root
                  is ``expr``
        """
        return AssertionError("unsupported expression type %s" % type(expr))

    @recurse.register(imp.Terminal)
    def recurse_terminal(expr, loop_indices):
        return temp_refcount(numbering, expr)

    @recurse.register(imp.For)
    def recurse_for(expr, loop_indices):
        return recurse(expr.children[0], loop_indices + (expr.index,))

    @recurse.register(imp.Block)
    def recurse_block(expr, loop_indices):
        # Temporaries declared at the beginning of the block are
        # collected here
        declare[expr] = []

        # Collect reference counts for the block
        refcount = collections.Counter()
        for statement in expr.children:
            refcount.update(recurse(statement, loop_indices))

        # Visit :class:`collections.Counter` in deterministic order
        for e in sorted(refcount.keys(), key=lambda t: numbering[t]):
            if refcount[e] == total_refcount[e]:
                # If all references are within this block, then this
                # block is the right place to declare the temporary.
                assert loop_indices == get_indices(e)[:len(loop_indices)]
                indices[e] = get_indices(e)[len(loop_indices):]
                if indices[e]:
                    # Scalar-valued temporaries are not declared until
                    # their value is assigned.  This does not really
                    # matter, but produces a more compact and nicer to
                    # read C code.
                    declare[expr].append(e)
                # Remove expression from the ``refcount`` so it will
                # not be declared again.
                del refcount[e]
        return refcount

    # Populate result
    remainder = recurse(tree, ())
    assert not remainder

    # Set in ``declare`` for Impero terminals whether they should
    # declare the temporary that they are writing to.
    for node in traversal((tree,)):
        if isinstance(node, imp.Terminal):
            declare[node] = False
            if isinstance(node, imp.Evaluate):
                e = node.expression
            elif isinstance(node, imp.Initialise):
                e = node.indexsum
            else:
                continue

            if len(indices[e]) == 0:
                declare[node] = True

    return declare, indices


def temp_refcount(temporaries, op):
    """Collects the number of times temporaries are referenced when
    generating code for an Impero terminal.

    :arg temporaries: set of temporaries
    :arg op: Impero terminal
    :returns: :class:`collections.Counter` object mapping some of
               elements from ``temporaries`` to the number of times
               they will referenced from ``op``
    """
    counter = collections.Counter()

    def recurse(o):
        """Traverses expression until reaching temporaries, counting
        temporary references."""
        if o in temporaries:
            counter[o] += 1
        else:
            for c in o.children:
                recurse(c)

    def recurse_top(o):
        """Traverses expression until reaching temporaries, counting
        temporary references. Always descends into children at least
        once, even when the root is a temporary."""
        if o in temporaries:
            counter[o] += 1
        for c in o.children:
            recurse(c)

    if isinstance(op, imp.Initialise):
        counter[op.indexsum] += 1
    elif isinstance(op, imp.Accumulate):
        recurse_top(op.indexsum)
    elif isinstance(op, imp.Evaluate):
        recurse_top(op.expression)
    elif isinstance(op, imp.Return):
        recurse(op.expression)
    elif isinstance(op, imp.ReturnAccumulate):
        recurse(op.indexsum.children[0])
    elif isinstance(op, imp.Noop):
        pass
    else:
        raise AssertionError("unhandled operation: %s" % type(op))

    return counter



from FIAT.polynomial_set import mis
from FIAT.reference_element import LINE

import gem
from gem.utils import cached_property

from finat.finiteelementbase import FiniteElementBase


class RuntimeTabulated(FiniteElementBase):
    """Element placeholder for tabulations provided at run time through a
    kernel argument.

    Used by Themis.
    """

    def __init__(self, cell, degree, variant=None, shift_axes=0,
                 restriction=None, continuous=True):
        """Construct a runtime tabulated element.

        :arg cell: reference cell
        :arg degree: polynomial degree (int)
        :arg variant: variant string of the UFL element
        :arg shift_axes: first dimension
        :arg restriction: None for single-cell integrals, '+' or '-'
                          for interior facet integrals depending on
                          which we need the tabulation on
        :arg continuous: continuous or discontinuous element?
        """
        # Currently only interval elements are accepted.
        if cell.get_shape() != LINE:
            raise NotImplementedError("Runtime tabulated elements limited to 1D.")

        # Sanity check
        assert isinstance(variant, str)
        assert isinstance(shift_axes, int) and 0 <= shift_axes
        assert isinstance(continuous, bool)
        assert restriction in [None, '+', '-']

        self.cell = cell
        self.degree = degree
        self.variant = variant
        self.shift_axes = shift_axes
        self.restriction = restriction
        self.continuous = continuous

    @cached_property
    def cell(self):
        pass  # set at initialization

    @cached_property
    def degree(self):
        pass  # set at initialization

    @cached_property
    def formdegree(self):
        if self.continuous:
            return 0
        else:
            return self.cell.get_spatial_dimension()

    def entity_dofs(self):
        raise NotImplementedError("I cannot tell where my DoFs are... :-/")

    def space_dimension(self):
        return self.degree + 1

    def basis_evaluation(self, order, ps, entity=None, coordinate_mapping=None):
        """Return code for evaluating the element at known points on the
        reference element.

        :param order: return derivatives up to this order.
        :param ps: the point set object.
        :param entity: the cell entity on which to tabulate.
        """
        # Spatial dimension
        dimension = self.cell.get_spatial_dimension()

        # Shape of the tabulation matrix
        shape = tuple(index.extent for index in ps.indices) + self.index_shape + self.value_shape

        result = {}
        for derivative in range(order + 1):
            for alpha in mis(dimension, derivative):
                name = str.format("rt_{}_{}_{}_{}_{}_{}",
                                  self.variant,
                                  self.degree,
                                  ''.join(map(str, alpha)),
                                  self.shift_axes,
                                  'c' if self.continuous else 'd',
                                  {None: "",
                                   '+': "p",
                                   '-': "m"}[self.restriction])
                result[alpha] = gem.partial_indexed(gem.Variable(name, shape), ps.indices)
        return result

    def point_evaluation(self, order, point, entity=None, coordinate_mapping=None):
        raise NotImplementedError("Point evaluation not supported for runtime tabulated elements")

    @property
    def index_shape(self):
        return (self.space_dimension(),)

    @property
    def value_shape(self):
        return ()

    @property
    def mapping(self):
        return "affine"



import numpy as np
from FIAT import ufc_simplex, ArnoldWintherNC, make_quadrature, expansions


def test_dofs():
    line = ufc_simplex(1)
    T = ufc_simplex(2)
    T.vertices = ((0.0, 0.0), (1.0, 0.25), (-0.75, 1.1))
    AW = ArnoldWintherNC(T, 2)

    Qline = make_quadrature(line, 6)

    linebfs = expansions.LineExpansionSet(line)
    linevals = linebfs.tabulate(1, Qline.pts)

    # n, n moments
    for ed in range(3):
        n = T.compute_scaled_normal(ed)
        wts = np.asarray(Qline.wts)
        nqpline = len(wts)

        vals = AW.tabulate(0, Qline.pts, (1, ed))[(0, 0)]
        nnvals = np.zeros((18, nqpline))
        for i in range(18):
            for j in range(len(wts)):
                nnvals[i, j] = n @ vals[i, :, :, j] @ n

        nnmoments = np.zeros((18, 2))

        for bf in range(18):
            for k in range(nqpline):
                for m in (0, 1):
                    nnmoments[bf, m] += wts[k] * nnvals[bf, k] * linevals[m, k]

        for bf in range(18):
            if bf != AW.dual.entity_ids[1][ed][0] and bf != AW.dual.entity_ids[1][ed][2]:
                assert np.allclose(nnmoments[bf, :], np.zeros(2))

    # n, t moments
    for ed in range(3):
        n = T.compute_scaled_normal(ed)
        t = T.compute_edge_tangent(ed)
        wts = np.asarray(Qline.wts)
        nqpline = len(wts)

        vals = AW.tabulate(0, Qline.pts, (1, ed))[(0, 0)]
        ntvals = np.zeros((18, nqpline))
        for i in range(18):
            for j in range(len(wts)):
                ntvals[i, j] = n @ vals[i, :, :, j] @ t

        ntmoments = np.zeros((18, 2))

        for bf in range(18):
            for k in range(nqpline):
                for m in (0, 1):
                    ntmoments[bf, m] += wts[k] * ntvals[bf, k] * linevals[m, k]

        for bf in range(18):
            if bf != AW.dual.entity_ids[1][ed][1] and bf != AW.dual.entity_ids[1][ed][3]:
                assert np.allclose(ntmoments[bf, :], np.zeros(2), atol=1.e-7)

    # check internal dofs
    ns = list(map(T.compute_scaled_normal, range(3)))
    Q = make_quadrature(T, 6)
    qpvals = AW.tabulate(0, Q.pts)[(0, 0)]
    const_moms = qpvals @ Q.wts / T.volume()
    nn_moms = const_moms.copy()
    for j in range(2):
        for i in range(2):
            comp = np.outer(ns[i+1], ns[j+1])
            nn_moms[:, i, j] = np.tensordot(const_moms, comp, ((1, 2), (0, 1)))
    assert np.allclose(nn_moms[:12], np.zeros((12, 2, 2)))
    assert np.allclose(nn_moms[15:], np.zeros((3, 2, 2)))
    assert np.allclose(nn_moms[12:15, 0, 0], np.asarray([1, 0, 0]))
    assert np.allclose(nn_moms[12:15, 0, 1], np.asarray([0, 1, 0]))
    assert np.allclose(nn_moms[12:15, 1, 0], np.asarray([0, 1, 0]))
    assert np.allclose(nn_moms[12:15, 1, 1], np.asarray([0, 0, 1]))



import pytest

import FIAT
import finat
import gem
import numpy


@pytest.fixture(params=[1, 2, 3])
def cell(request):
    dim = request.param
    return FIAT.ufc_simplex(dim)


@pytest.mark.parametrize('degree', [1, 2])
def test_cellwise_constant(cell, degree):
    dim = cell.get_spatial_dimension()
    element = finat.Lagrange(cell, degree)
    index = gem.Index()
    point = gem.partial_indexed(gem.Variable('X', (17, dim)), (index,))

    order = 2
    for alpha, table in element.point_evaluation(order, point).items():
        if sum(alpha) < degree:
            assert table.free_indices == (index,)
        else:
            assert table.free_indices == ()


@pytest.mark.parametrize("element,degree", [
    (finat.HsiehCloughTocher, 3),
    (finat.Argyris, 5),
    (finat.MardalTaiWinther, 3),
])
def test_point_evaluation_zany(ref_to_phys, element, degree):
    dim = 2
    ref_cell = ref_to_phys[dim].ref_cell
    phys_cell = ref_to_phys[dim].phys_cell
    A, b = FIAT.reference_element.make_affine_mapping(ref_cell.vertices, phys_cell.vertices)

    ref_pt = numpy.array([0.2, 0.3])
    phys_pt = numpy.dot(A, ref_pt) + b

    finat_kwargs = {}
    if element in {finat.HsiehCloughTocher, finat.Argyris}:
        finat_kwargs["avg"] = True

    order = 0
    point = gem.Literal(ref_pt)
    ref_element = element(ref_cell, degree, **finat_kwargs)
    result = ref_element.point_evaluation(order, point, coordinate_mapping=ref_to_phys[dim])

    phys_element = element(phys_cell, degree, **finat_kwargs).fiat_equivalent
    expected = phys_element.tabulate(order, phys_pt)

    num_dof = ref_element.space_dimension()
    for alpha in result:

        ref_val, = gem.interpreter.evaluate([result[alpha]])

        if phys_element.mapping()[0] == "covariant piola":
            val = numpy.tensordot(ref_val.arr, A, (-1, 0))
        elif phys_element.mapping()[0] == "contravariant piola":
            detA = numpy.linalg.det(A)
            val = numpy.tensordot(ref_val.arr, A/detA, (-1, 1))
        else:
            val = ref_val.arr

        assert numpy.allclose(val, expected[alpha][:num_dof])


if __name__ == '__main__':
    import os
    pytest.main(os.path.abspath(__file__))



# Copyright (C) 2015 Imperial College London and others.
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Written by Pablo D. Brubeck (brubeck@protonmail.com), 2022

import numpy

from FIAT import finite_element, dual_set, functional, P0
from FIAT.reference_element import symmetric_simplex
from FIAT.orientation_utils import make_entity_permutations_simplex
from FIAT.quadrature import FacetQuadratureRule
from FIAT.quadrature_schemes import create_quadrature
from FIAT.polynomial_set import ONPolynomialSet, make_bubbles
from FIAT.check_format_variant import parse_lagrange_variant


def make_dual_bubbles(ref_el, degree, codim=0, interpolant_deg=None):
    """Tabulate the L2-duals of the hierarchical C0 basis."""
    if interpolant_deg is None:
        interpolant_deg = degree
    Q = create_quadrature(ref_el, degree + interpolant_deg)
    B = make_bubbles(ref_el, degree, codim=codim, scale="orthonormal")
    P_at_qpts = B.expansion_set.tabulate(degree, Q.get_points())
    M = numpy.dot(numpy.multiply(P_at_qpts, Q.get_weights()), P_at_qpts.T)
    phis = numpy.linalg.solve(M, P_at_qpts)
    phis = numpy.dot(B.get_coeffs(), phis)
    return Q, phis


class LegendreDual(dual_set.DualSet):
    """The dual basis for Legendre elements."""
    def __init__(self, ref_el, degree, codim=0):
        nodes = []
        entity_ids = {}
        entity_permutations = {}

        sd = ref_el.get_spatial_dimension()
        top = ref_el.get_topology()
        for dim in sorted(top):
            npoints = degree + 1 if dim == sd - codim else 0
            perms = make_entity_permutations_simplex(dim, npoints)
            entity_permutations[dim] = {}
            entity_ids[dim] = {}
            if npoints == 0:
                for entity in sorted(top[dim]):
                    entity_ids[dim][entity] = []
                    entity_permutations[dim][entity] = perms
                continue

            ref_facet = ref_el.construct_subelement(dim)
            poly_set = ONPolynomialSet(ref_facet, degree)
            Q_ref = create_quadrature(ref_facet, 2 * degree)
            phis = poly_set.tabulate(Q_ref.get_points())[(0,) * dim]
            for entity in sorted(top[dim]):
                cur = len(nodes)
                Q_facet = FacetQuadratureRule(ref_el, dim, entity, Q_ref)
                nodes.extend(functional.IntegralMoment(ref_el, Q_facet, phi) for phi in phis)
                entity_ids[dim][entity] = list(range(cur, len(nodes)))
                entity_permutations[dim][entity] = perms

        super().__init__(nodes, ref_el, entity_ids, entity_permutations)


class Legendre(finite_element.CiarletElement):
    """Simplicial discontinuous element with Legendre polynomials."""
    def __new__(cls, ref_el, degree, variant=None):
        if degree == 0:
            splitting, _ = parse_lagrange_variant(variant, integral=True)
            if splitting is None:
                # FIXME P0 on the split requires implementing SplitSimplicialComplex.symmetry_group_size()
                return P0.P0(ref_el)
        return super().__new__(cls)

    def __init__(self, ref_el, degree, variant=None):
        splitting, _ = parse_lagrange_variant(variant, integral=True)
        if splitting is not None:
            ref_el = splitting(ref_el)
        poly_set = ONPolynomialSet(ref_el, degree)
        dual = LegendreDual(ref_el, degree)
        formdegree = ref_el.get_spatial_dimension()  # n-form
        super().__init__(poly_set, dual, degree, formdegree)


class IntegratedLegendreDual(dual_set.DualSet):
    """The dual basis for integrated Legendre elements."""
    def __init__(self, ref_el, degree):
        nodes = []
        entity_ids = {}
        entity_permutations = {}

        top = ref_el.get_topology()
        for dim in sorted(top):
            perms = make_entity_permutations_simplex(dim, degree - dim)
            entity_ids[dim] = {}
            entity_permutations[dim] = {}
            if dim == 0 or degree <= dim:
                for entity in sorted(top[dim]):
                    cur = len(nodes)
                    pts = ref_el.make_points(dim, entity, degree)
                    nodes.extend(functional.PointEvaluation(ref_el, pt) for pt in pts)
                    entity_ids[dim][entity] = list(range(cur, len(nodes)))
                    entity_permutations[dim][entity] = perms
                continue

            ref_facet = symmetric_simplex(dim)
            Q_ref, phis = make_dual_bubbles(ref_facet, degree)
            for entity in sorted(top[dim]):
                cur = len(nodes)
                Q_facet = FacetQuadratureRule(ref_el, dim, entity, Q_ref)

                # phis must transform like a d-form to undo the measure transformation
                scale = 1 / Q_facet.jacobian_determinant()
                Jphis = scale * phis

                nodes.extend(functional.IntegralMoment(ref_el, Q_facet, phi) for phi in Jphis)
                entity_ids[dim][entity] = list(range(cur, len(nodes)))
                entity_permutations[dim][entity] = perms

        super().__init__(nodes, ref_el, entity_ids, entity_permutations)


class IntegratedLegendre(finite_element.CiarletElement):
    """Simplicial continuous element with integrated Legendre polynomials."""
    def __init__(self, ref_el, degree, variant=None):
        splitting, _ = parse_lagrange_variant(variant, integral=True)
        if splitting is not None:
            ref_el = splitting(ref_el)
        if degree < 1:
            raise ValueError(f"{type(self).__name__} elements only valid for k >= 1")
        poly_set = ONPolynomialSet(ref_el, degree, variant="bubble")
        dual = IntegratedLegendreDual(ref_el, degree)
        formdegree = 0  # 0-form
        super().__init__(poly_set, dual, degree, formdegree)



# Copyright (C) 2015 Imperial College London and others.
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Written by David A. Ham (david.ham@imperial.ac.uk), 2015
#
# Modified by Pablo D. Brubeck (brubeck@protonmail.com), 2021

from FIAT import discontinuous_lagrange


class GaussLegendre(discontinuous_lagrange.DiscontinuousLagrange):
    """Simplicial discontinuous element with nodes at the (recursive) Gauss-Legendre points."""
    def __init__(self, ref_el, degree):
        super().__init__(ref_el, degree, variant="gl")



# Copyright (C) 2008 Robert C. Kirby (Texas Tech University)
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Marie E. Rognes (meg@simula.no), 2012
# Modified by David A. Ham (david.ham@imperial.ac.uk), 2015

import itertools
from math import factorial
import numpy
from recursivenodes.quadrature import gaussjacobi, lobattogaussjacobi, simplexgausslegendre

from FIAT import reference_element
from FIAT.orientation_utils import make_entity_permutations_simplex


def pseudo_determinant(A):
    return numpy.sqrt(abs(numpy.linalg.det(numpy.dot(A.T, A))))


def map_quadrature(pts_ref, wts_ref, source_cell, target_cell, jacobian=False):
    """Map quadrature points and weights defined on source_cell to target_cell.
    """
    while source_cell.get_parent():
        source_cell = source_cell.get_parent()
    A, b = reference_element.make_affine_mapping(source_cell.get_vertices(),
                                                 target_cell.get_vertices())
    if len(pts_ref.shape) != 2:
        pts_ref = pts_ref.reshape(-1, A.shape[1])
    scale = pseudo_determinant(A)
    pts = numpy.dot(pts_ref, A.T)
    pts = numpy.add(pts, b, out=pts)
    wts = scale * wts_ref

    # return immutable types
    pts = tuple(map(tuple, pts))
    wts = tuple(wts.flat)
    if jacobian:
        return pts, wts, A
    return pts, wts


class QuadratureRule(object):
    """General class that models integration over a reference element
    as the weighted sum of a function evaluated at a set of points.
    """
    def __init__(self, ref_el, pts, wts):
        if len(wts) != len(pts):
            raise ValueError("Have %d weights, but %d points" % (len(wts), len(pts)))

        self.ref_el = ref_el
        self.pts = pts
        self.wts = wts
        self._intrinsic_orientation_permutation_map_tuple = (None, )

    def get_points(self):
        return numpy.array(self.pts)

    def get_weights(self):
        return numpy.array(self.wts)

    def integrate(self, f):
        return sum(w * f(x) for x, w in zip(self.pts, self.wts))

    @property
    def extrinsic_orientation_permutation_map(self):
        """A map from extrinsic orientations to corresponding axis permutation matrices.

        Notes
        -----
        result[eo] gives the physical axis-reference axis permutation matrix corresponding to
        eo (extrinsic orientation).

        """
        return self.ref_el.extrinsic_orientation_permutation_map

    @property
    def intrinsic_orientation_permutation_map_tuple(self):
        """A tuple of maps from intrinsic orientations to corresponding point permutations for each reference cell axis.

        Notes
        -----
        result[axis][io] gives the physical point-reference point permutation array corresponding to
        io (intrinsic orientation) on ``axis``.

        """
        if any(m is None for m in self._intrinsic_orientation_permutation_map_tuple):
            raise ValueError("Must set _intrinsic_orientation_permutation_map_tuple")
        return self._intrinsic_orientation_permutation_map_tuple


class GaussJacobiQuadratureLineRule(QuadratureRule):
    """Gauss-Jacobi quadature rule determined by Jacobi weights a and b
    using m roots of m:th order Jacobi polynomial."""

    def __init__(self, ref_el, m, a=0, b=0):
        Ref1 = reference_element.DefaultLine()
        pts_ref, wts_ref = gaussjacobi(m, a, b)
        pts, wts = map_quadrature(pts_ref, wts_ref, Ref1, ref_el)
        QuadratureRule.__init__(self, ref_el, pts, wts)
        # Set _intrinsic_orientation_permutation_map_tuple.
        dim = 1
        a = numpy.zeros((factorial(dim + 1), m), dtype=int)
        for io, perm in make_entity_permutations_simplex(dim, m).items():
            a[io, perm] = range(m)
        self._intrinsic_orientation_permutation_map_tuple = (a, )


class GaussLobattoLegendreQuadratureLineRule(QuadratureRule):
    """Gauss-Lobatto-Legendre quadrature rule on the interval.

    The quadrature rule uses m points for a degree of precision of 2m-3.
    """
    def __init__(self, ref_el, m):
        if m < 2:
            raise ValueError(
                "Gauss-Labotto-Legendre quadrature invalid for fewer than 2 points")
        Ref1 = reference_element.DefaultLine()
        pts_ref, wts_ref = lobattogaussjacobi(m, 0, 0)
        pts, wts = map_quadrature(pts_ref, wts_ref, Ref1, ref_el)
        QuadratureRule.__init__(self, ref_el, pts, wts)


class GaussLegendreQuadratureLineRule(GaussJacobiQuadratureLineRule):
    """Gauss--Legendre quadrature rule on the interval.

    The quadrature rule uses m points for a degree of precision of 2m-1.
    """
    def __init__(self, ref_el, m):
        super().__init__(ref_el, m)


class RadauQuadratureLineRule(QuadratureRule):
    """Gauss--Radau quadrature rule on the interval.

    The quadrature rule uses m points for a degree of precision of 2m-1.
    """
    def __init__(self, ref_el, m, right=True):
        if m < 1:
            raise ValueError(
                "Gauss-Radau quadrature invalid for fewer than 1 points")

        right = int(right)
        x0 = ref_el.vertices[right]
        volume = ref_el.volume()
        if m > 1:
            # Make the interior points and weights
            rule = GaussJacobiQuadratureLineRule(ref_el, m-1, right, 1-right)
            # Remove the hat weight from the quadrature weights
            x = rule.get_points().reshape((-1,))
            hat = (2.0 / volume) * abs(x0[0] - x)
            wts = rule.get_weights() / hat
            pts = rule.pts
        else:
            # Special case for lowest order.
            wts = ()
            pts = ()

        # Get the weight at the endpoint via sum(ws) == volume
        w0 = volume - sum(wts)
        xs = (*pts, x0) if right else (x0, *pts)
        ws = (*wts, w0) if right else (w0, *wts)

        QuadratureRule.__init__(self, ref_el, xs, ws)


class CollapsedQuadratureSimplexRule(QuadratureRule):
    """Implements the collapsed quadrature rules defined in
    Karniadakis & Sherwin by mapping products of Gauss-Jacobi rules
    from the hypercube to the simplex."""

    def __init__(self, ref_el, m):
        dim = ref_el.get_spatial_dimension()
        Ref1 = reference_element.default_simplex(dim)
        pts_ref, wts_ref = simplexgausslegendre(dim, m)
        pts, wts = map_quadrature(pts_ref, wts_ref, Ref1, ref_el)
        QuadratureRule.__init__(self, ref_el, pts, wts)


class CollapsedQuadratureTriangleRule(CollapsedQuadratureSimplexRule):
    """Implements the collapsed quadrature rules defined in
    Karniadakis & Sherwin by mapping products of Gauss-Jacobi rules
    from the square to the triangle."""
    pass


class CollapsedQuadratureTetrahedronRule(CollapsedQuadratureSimplexRule):
    """Implements the collapsed quadrature rules defined in
    Karniadakis & Sherwin by mapping products of Gauss-Jacobi rules
    from the cube to the tetrahedron."""
    pass


class FacetQuadratureRule(QuadratureRule):
    """A quadrature rule on a facet mapped from a reference quadrature rule.
    """
    def __init__(self, ref_el, entity_dim, entity_id, Q_ref):
        # Construct the facet of interest
        facet = ref_el.construct_subelement(entity_dim)
        facet_topology = ref_el.get_topology()[entity_dim][entity_id]
        facet.vertices = ref_el.get_vertices_of_subcomplex(facet_topology)

        # Map reference points and weights on the appropriate facet
        pts_ref = Q_ref.get_points()
        wts_ref = Q_ref.get_weights()
        pts, wts, J = map_quadrature(pts_ref, wts_ref, Q_ref.ref_el, facet, jacobian=True)

        # Initialize super class with new points and weights
        QuadratureRule.__init__(self, facet, pts, wts)
        self._J = J
        self._reference_rule = Q_ref

    def reference_rule(self):
        return self._reference_rule

    def jacobian(self):
        return self._J

    def jacobian_determinant(self):
        return pseudo_determinant(self._J)


def make_quadrature(ref_el, m):
    """Returns the collapsed quadrature rule using m points per
    direction on the given reference element. In the tensor product
    case, m is a tuple."""

    if isinstance(m, tuple):
        min_m = min(m)
    else:
        min_m = m

    msg = "Expecting at least one (not %d) quadrature point per direction" % min_m
    assert (min_m > 0), msg

    if ref_el.get_shape() == reference_element.POINT:
        return QuadratureRule(ref_el, [()], [1])
    elif ref_el.get_shape() == reference_element.LINE:
        return GaussJacobiQuadratureLineRule(ref_el, m)
    elif ref_el.get_shape() == reference_element.TRIANGLE:
        return CollapsedQuadratureTriangleRule(ref_el, m)
    elif ref_el.get_shape() == reference_element.TETRAHEDRON:
        return CollapsedQuadratureTetrahedronRule(ref_el, m)
    elif ref_el.get_shape() == reference_element.QUADRILATERAL:
        line_rule = GaussJacobiQuadratureLineRule(ref_el.construct_subelement(1), m)
        return make_tensor_product_quadrature(line_rule, line_rule)
    elif ref_el.get_shape() == reference_element.HEXAHEDRON:
        line_rule = GaussJacobiQuadratureLineRule(ref_el.construct_subelement(1), m)
        return make_tensor_product_quadrature(line_rule, line_rule, line_rule)
    else:
        raise ValueError("Unable to make quadrature for cell: %s" % ref_el)


def make_tensor_product_quadrature(*quad_rules):
    """Returns the quadrature rule for a TensorProduct cell, by combining
    the quadrature rules of the components."""
    ref_el = reference_element.TensorProductCell(*[q.ref_el
                                                   for q in quad_rules])
    # Coordinates are "concatenated", weights are multiplied
    pts = [list(itertools.chain(*pt_tuple))
           for pt_tuple in itertools.product(*[q.pts for q in quad_rules])]
    wts = [numpy.prod(wt_tuple)
           for wt_tuple in itertools.product(*[q.wts for q in quad_rules])]
    return QuadratureRule(ref_el, pts, wts)



import FIAT
import numpy
from gem import ListTensor

from finat.fiat_elements import FiatElement
from finat.physically_mapped import Citations, identity, PhysicallyMappedElement
from finat.piola_mapped import piola_inverse


class AlfeldSorokina(PhysicallyMappedElement, FiatElement):
    def __init__(self, cell, degree=2):
        if Citations is not None:
            Citations().register("AlfeldSorokina2016")
        super().__init__(FIAT.AlfeldSorokina(cell, degree))

    def basis_transformation(self, coordinate_mapping):
        sd = self.cell.get_spatial_dimension()
        bary, = self.cell.make_points(sd, 0, sd+1)
        J = coordinate_mapping.jacobian_at(bary)
        detJ = coordinate_mapping.detJ_at(bary)

        dofs = self.entity_dofs()
        V = identity(self.space_dimension())

        # Undo the Piola transform
        nodes = self._element.get_dual_set().nodes
        Finv = piola_inverse(self.cell, J, detJ)
        for dim in sorted(dofs):
            for e in sorted(dofs[dim]):
                k = 0
                while k < len(dofs[dim][e]):
                    cur = dofs[dim][e][k]
                    if len(nodes[cur].deriv_dict) > 0:
                        V[cur, cur] = detJ
                        k += 1
                    else:
                        s = dofs[dim][e][k:k+sd]
                        V[numpy.ix_(s, s)] = Finv
                        k += sd

        return ListTensor(V.T)



import FIAT

from finat.physically_mapped import Citations
from finat.fiat_elements import FiatElement
from finat.piola_mapped import PiolaBubbleElement


class ArnoldQin(FiatElement):
    def __init__(self, cell, degree=2):
        if Citations is not None:
            Citations().register("ArnoldQin1992")
        super().__init__(FIAT.ArnoldQin(cell, degree))


class ReducedArnoldQin(PiolaBubbleElement):
    def __init__(self, cell, degree=2):
        if Citations is not None:
            Citations().register("ArnoldQin1992")
        super().__init__(FIAT.ArnoldQin(cell, degree, reduced=True))



import FIAT

from finat.physically_mapped import Citations
from finat.piola_mapped import PiolaBubbleElement


class BernardiRaugel(PiolaBubbleElement):
    def __init__(self, cell, order=1):
        if Citations is not None:
            Citations().register("BernardiRaugel1985")
        super().__init__(FIAT.BernardiRaugel(cell, order=order))


class BernardiRaugelBubble(BernardiRaugel):
    def __init__(self, cell, degree=None):
        super().__init__(cell, order=0)



# Copyright (C) 2008 Robert C. Kirby (Texas Tech University)
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from FIAT import finite_element, polynomial_set, dual_set, functional
from FIAT.reference_element import TETRAHEDRON, TRIANGLE
from FIAT.quadrature import FacetQuadratureRule
from FIAT.quadrature_schemes import create_quadrature
import numpy
import math


class MorleyDualSet(dual_set.DualSet):
    """The dual basis for Morley elements.  This class works for
    simplices of any dimension.  Nodes are average on codim=2 entities
    and average of normal derivative on codim=1 entities."""

    def __init__(self, ref_el, degree):
        top = ref_el.get_topology()
        sd = ref_el.get_spatial_dimension()
        entity_ids = {dim: {entity: [] for entity in top[dim]} for dim in top}
        nodes = []

        def duals(ref_el, dim, degree):
            facet = ref_el.construct_subelement(dim)
            Q_ref = create_quadrature(facet, degree)
            scale = numpy.ones(Q_ref.get_weights().shape)
            return Q_ref, scale

        # codim=2 dof -- integral average
        dim = sd - 2
        Q_ref, scale = duals(ref_el, dim, degree)
        for entity in sorted(top[dim]):
            cur = len(nodes)
            Q = FacetQuadratureRule(ref_el, dim, entity, Q_ref)
            nodes.append(functional.IntegralMoment(ref_el, Q, scale / Q.jacobian_determinant()))
            entity_ids[dim][entity].extend(range(cur, len(nodes)))

        # codim=1 dof -- average of normal derivative at each facet
        dim = sd - 1
        Q_ref, scale = duals(ref_el, dim, degree-1)
        # normalized normals do not have unit norm!
        scale /= math.factorial(sd-1)
        for entity in sorted(top[dim]):
            cur = len(nodes)
            nodes.append(functional.IntegralMomentOfNormalDerivative(ref_el, entity, Q_ref, scale))
            entity_ids[dim][entity].extend(range(cur, len(nodes)))

        super().__init__(nodes, ref_el, entity_ids)


class Morley(finite_element.CiarletElement):
    """The Morley finite element."""

    def __init__(self, ref_el, degree=2):
        if ref_el.get_shape() not in {TRIANGLE, TETRAHEDRON}:
            raise ValueError("Morley only defined on simplices of dimension >= 2")
        if degree != 2:
            raise ValueError("{type(self).__name__} only defined for degree == 2")
        poly_set = polynomial_set.ONPolynomialSet(ref_el, degree)
        dual = MorleyDualSet(ref_el, degree)
        super().__init__(poly_set, dual, degree)



import finat
import numpy as np
import pytest
from gem.interpreter import evaluate


@pytest.mark.parametrize("element, degree, variant", [
    (finat.Hermite, 3, None),
    (finat.QuadraticPowellSabin6, 2, None),
    (finat.QuadraticPowellSabin12, 2, None),
    (finat.ReducedHsiehCloughTocher, 3, None),
    (finat.HsiehCloughTocher, 3, None),
    (finat.HsiehCloughTocher, 4, None),
    (finat.Bell, 5, None),
    (finat.Argyris, 5, "point"),
    (finat.Argyris, 5, None),
    (finat.Argyris, 6, None),
])
def test_mass_scaling(scaled_ref_to_phys, element, degree, variant):
    sd = 2
    ref_cell = scaled_ref_to_phys[sd][0].ref_cell
    if variant is not None:
        ref_element = element(ref_cell, degree, variant=variant)
    else:
        ref_element = element(ref_cell, degree)

    Q = finat.quadrature.make_quadrature(ref_cell, 2*degree)
    qpts = Q.point_set
    qwts = Q.weights

    kappa = []
    for mapping in scaled_ref_to_phys[sd]:
        J_gem = mapping.jacobian_at(ref_cell.make_points(sd, 0, sd+1)[0])
        J = evaluate([J_gem])[0].arr

        z = (0,) * ref_element.cell.get_spatial_dimension()
        finat_vals_gem = ref_element.basis_evaluation(0, qpts, coordinate_mapping=mapping)[z]
        phis = evaluate([finat_vals_gem])[0].arr.T

        M = np.dot(np.multiply(phis, qwts * abs(np.linalg.det(J))), phis.T)
        kappa.append(np.linalg.cond(M))

    kappa = np.array(kappa)
    ratio = kappa[1:] / kappa[:-1]
    assert np.allclose(ratio, 1, atol=0.1)



# -*- coding: utf-8 -*-
"""Implementation of the Hu-Zhang finite elements."""

# Copyright (C) 2024 by Francis Aznaran (University of Notre Dame)
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later


from FIAT import finite_element, polynomial_set, dual_set
from FIAT.check_format_variant import check_format_variant
from FIAT.reference_element import TRIANGLE
from FIAT.quadrature import FacetQuadratureRule
from FIAT.quadrature_schemes import create_quadrature
from FIAT.functional import (ComponentPointEvaluation,
                             PointwiseInnerProductEvaluation,
                             TensorBidirectionalIntegralMoment,
                             IntegralLegendreNormalNormalMoment,
                             IntegralLegendreNormalTangentialMoment)


class HuZhangDual(dual_set.DualSet):
    def __init__(self, ref_el, degree, variant, qdegree):
        top = ref_el.get_topology()
        sd = ref_el.get_spatial_dimension()
        shp = (sd, sd)
        entity_ids = {dim: {entity: [] for entity in sorted(top[dim])} for dim in sorted(top)}
        nodes = []

        # vertex dofs
        for v in sorted(top[0]):
            cur = len(nodes)
            pt, = ref_el.make_points(0, v, degree)
            nodes.extend(ComponentPointEvaluation(ref_el, (i, j), shp, pt)
                         for i in range(sd) for j in range(i, sd))
            entity_ids[0][v].extend(range(cur, len(nodes)))

        # edge dofs
        for entity in sorted(top[1]):
            cur = len(nodes)
            if variant == "point":
                # nn and nt components evaluated at edge points
                n = ref_el.compute_scaled_normal(entity)
                t = ref_el.compute_edge_tangent(entity)
                pts = ref_el.make_points(1, entity, degree)
                nodes.extend(PointwiseInnerProductEvaluation(ref_el, n, s, pt)
                             for pt in pts for s in (n, t))

            elif variant == "integral":
                # bidirectional nn and nt moments against P_{k-2}
                moments = (IntegralLegendreNormalNormalMoment, IntegralLegendreNormalTangentialMoment)
                nodes.extend(mu(ref_el, entity, order, qdegree + degree-2)
                             for order in range(degree-1) for mu in moments)
            entity_ids[1][entity].extend(range(cur, len(nodes)))

        # interior dofs
        if variant == "integral":
            cell = ref_el.construct_subelement(sd)
            Q_ref = create_quadrature(cell, 2*degree-2)
            P = polynomial_set.ONPolynomialSet(cell, degree-2, scale=1)
            Phis = P.tabulate(Q_ref.get_points())[(0,)*sd]

        for entity in sorted(top[sd]):
            cur = len(nodes)
            if variant == "point":
                # unique components evaluated at interior points
                pts = ref_el.make_points(sd, entity, degree+1)
                nodes.extend(ComponentPointEvaluation(ref_el, (i, j), shp, pt)
                             for pt in pts for i in range(sd) for j in range(i, sd))

            elif variant == "integral":
                # Moments of unique components against a basis for P_{k-2}
                faces = ref_el.get_connectivity()[(sd, sd-1)][entity]
                n = list(map(ref_el.compute_scaled_normal, faces))
                Q = FacetQuadratureRule(ref_el, sd, entity, Q_ref)
                phis = Phis / Q.jacobian_determinant()
                nodes.extend(TensorBidirectionalIntegralMoment(ref_el, n[i+1], n[j+1], Q, phi)
                             for phi in phis for i in range(sd) for j in range(i, sd))

            entity_ids[sd][entity].extend(range(cur, len(nodes)))
        super().__init__(nodes, ref_el, entity_ids)


class HuZhang(finite_element.CiarletElement):
    """The definition of the Hu-Zhang element."""
    def __init__(self, ref_el, degree=3, variant=None):
        if degree < 3:
            raise ValueError(f"{type(self).__name__} only defined for degree >= 3")
        if ref_el.shape != TRIANGLE:
            raise ValueError(f"{type(self).__name__} only defined on triangles")
        splitting, variant, qdegree = check_format_variant(variant, degree)
        if splitting is not None:
            raise NotImplementedError(f"{type(self).__name__} is not implemented as a macroelement.")

        poly_set = polynomial_set.ONSymTensorPolynomialSet(ref_el, degree)
        dual = HuZhangDual(ref_el, degree, variant, qdegree)
        formdegree = ref_el.get_spatial_dimension() - 1
        mapping = "double contravariant piola"
        super().__init__(poly_set, dual, degree, formdegree, mapping=mapping)



"""
An interpreter for GEM trees.
"""
import numpy
import operator
from collections import OrderedDict
from functools import singledispatch
import itertools

from gem import gem, node
from gem.optimise import replace_delta

__all__ = ("evaluate", )


class Result(object):
    """An array object that tracks which axes of the array correspond to
    gem free indices (and what those free indices are).

    :arg arr: The array.
    :arg fids: The free indices.

    The first ``len(fids)`` axes of the provided array correspond to
    the free indices, the remaining axes are the shape of each entry.
    """
    def __init__(self, arr, fids=None):
        self.arr = arr
        self.fids = fids if fids is not None else ()

    def broadcast(self, fids):
        """Given some free indices, return a broadcasted array which
        contains extra dimensions that correspond to indices in fids
        that are not in ``self.fids``.

        Note that inserted dimensions will have length one.

        :arg fids: The free indices for broadcasting.
        """
        # Select free indices
        axes = tuple(self.fids.index(fi) for fi in fids if fi in self.fids)
        assert len(axes) == len(self.fids)
        # Add shape
        axes += tuple(range(len(self.fids), self.arr.ndim))
        # Move axes, insert extra axes
        arr = numpy.transpose(self.arr, axes)
        for i, fi in enumerate(fids):
            if fi not in self.fids:
                arr = numpy.expand_dims(arr, axis=i)
        return arr

    def filter(self, idx, fids):
        """Given an index tuple and some free indices, return a
        "filtered" index tuple which removes entries that correspond
        to indices in fids that are not in ``self.fids``.

        :arg idx: The index tuple to filter.
        :arg fids: The free indices for the index tuple.
        """
        return tuple(idx[fids.index(i)] for i in self.fids) + idx[len(fids):]

    def __getitem__(self, idx):
        return self.arr[tuple(idx)]

    def __setitem__(self, idx, val):
        self.arr[idx] = val

    @property
    def tshape(self):
        """The total shape of the result array."""
        return self.arr.shape

    @property
    def fshape(self):
        """The shape of the free index part of the result array."""
        return self.tshape[:len(self.fids)]

    @property
    def shape(self):
        """The shape of the shape part of the result array."""
        return self.tshape[len(self.fids):]

    def __repr__(self):
        return "Result(%r, %r)" % (self.arr, self.fids)

    def __str__(self):
        return repr(self)

    @classmethod
    def empty(cls, *children, **kwargs):
        """Build an empty Result object.

        :arg children: The children used to determine the shape and
            free indices.
        :kwarg dtype: The data type of the result array.
        """
        dtype = kwargs.get("dtype", float)
        assert all(children[0].shape == c.shape for c in children)
        fids = []
        for f in itertools.chain(*(c.fids for c in children)):
            if f not in fids:
                fids.append(f)
        shape = tuple(i.extent for i in fids) + children[0].shape
        return cls(numpy.empty(shape, dtype=dtype), tuple(fids))


@singledispatch
def _evaluate(expression, self):
    """Evaluate an expression using a provided callback handler.

    :arg expression: The expression to evaluation.
    :arg self: The callback handler (should provide bindings).
    """
    raise ValueError("Unhandled node type %s" % type(expression))


@_evaluate.register(gem.Zero)
def _evaluate_zero(e, self):
    """Zeros produce an array of zeros."""
    return Result(numpy.zeros(e.shape, dtype=float))


@_evaluate.register(gem.Failure)
def _evaluate_failure(e, self):
    """Failure nodes produce NaNs."""
    return Result(numpy.full(e.shape, numpy.nan, dtype=float))


@_evaluate.register(gem.Constant)
def _evaluate_constant(e, self):
    """Constants return their array."""
    return Result(e.array)


@_evaluate.register(gem.Delta)
def _evaluate_delta(e, self):
    """Lower delta and evaluate."""
    e, = replace_delta((e,))
    return self(e)


@_evaluate.register(gem.Variable)
def _evaluate_variable(e, self):
    """Look up variables in the provided bindings."""
    try:
        val = self.bindings[e]
    except KeyError:
        raise ValueError("Binding for %s not found" % e)
    if val.shape != e.shape:
        raise ValueError("Binding for %s has wrong shape.  %s, not %s." %
                         (e, val.shape, e.shape))
    return Result(val)


@_evaluate.register(gem.Power)
@_evaluate.register(gem.Division)
@_evaluate.register(gem.Product)
@_evaluate.register(gem.Sum)
def _evaluate_operator(e, self):
    op = {gem.Product: operator.mul,
          gem.Division: operator.truediv,
          gem.Sum: operator.add,
          gem.Power: operator.pow}[type(e)]

    a, b = [self(o) for o in e.children]
    result = Result.empty(a, b)
    fids = result.fids
    result.arr = op(a.broadcast(fids), b.broadcast(fids))
    return result


@_evaluate.register(gem.MathFunction)
def _evaluate_mathfunction(e, self):
    ops = [self(o) for o in e.children]
    result = Result.empty(*ops)
    names = {
        "abs": abs,
        "log": numpy.log,
        "real": operator.attrgetter("real"),
        "imag": operator.attrgetter("imag"),
        "conj": operator.methodcaller("conjugate"),
    }
    op = names[e.name]
    for idx in numpy.ndindex(result.tshape):
        result[idx] = op(*(o[o.filter(idx, result.fids)] for o in ops))
    return result


@_evaluate.register(gem.MaxValue)
@_evaluate.register(gem.MinValue)
def _evaluate_minmaxvalue(e, self):
    ops = [self(o) for o in e.children]
    result = Result.empty(*ops)
    op = {gem.MinValue: min,
          gem.MaxValue: max}[type(e)]
    for idx in numpy.ndindex(result.tshape):
        result[idx] = op(*(o[o.filter(idx, result.fids)] for o in ops))
    return result


@_evaluate.register(gem.Comparison)
def _evaluate_comparison(e, self):
    ops = [self(o) for o in e.children]
    op = {">": operator.gt,
          ">=": operator.ge,
          "==": operator.eq,
          "!=": operator.ne,
          "<": operator.lt,
          "<=": operator.le}[e.operator]
    result = Result.empty(*ops, dtype=bool)
    for idx in numpy.ndindex(result.tshape):
        result[idx] = op(*(o[o.filter(idx, result.fids)] for o in ops))
    return result


@_evaluate.register(gem.LogicalNot)
def _evaluate_logicalnot(e, self):
    val = self(e.children[0])
    assert val.arr.dtype == numpy.dtype("bool")
    result = Result.empty(val, bool)
    for idx in numpy.ndindex(result.tshape):
        result[idx] = not val[val.filter(idx, result.fids)]
    return result


@_evaluate.register(gem.LogicalAnd)
def _evaluate_logicaland(e, self):
    a, b = [self(o) for o in e.children]
    assert a.arr.dtype == numpy.dtype("bool")
    assert b.arr.dtype == numpy.dtype("bool")
    result = Result.empty(a, b, bool)
    for idx in numpy.ndindex(result.tshape):
        result[idx] = a[a.filter(idx, result.fids)] and \
            b[b.filter(idx, result.fids)]
    return result


@_evaluate.register(gem.LogicalOr)
def _evaluate_logicalor(e, self):
    a, b = [self(o) for o in e.children]
    assert a.arr.dtype == numpy.dtype("bool")
    assert b.arr.dtype == numpy.dtype("bool")
    result = Result.empty(a, b, dtype=bool)
    for idx in numpy.ndindex(result.tshape):
        result[idx] = a[a.filter(idx, result.fids)] or \
            b[b.filter(idx, result.fids)]
    return result


@_evaluate.register(gem.Conditional)
def _evaluate_conditional(e, self):
    cond, then, else_ = [self(o) for o in e.children]
    assert cond.arr.dtype == numpy.dtype("bool")
    result = Result.empty(cond, then, else_)
    for idx in numpy.ndindex(result.tshape):
        if cond[cond.filter(idx, result.fids)]:
            result[idx] = then[then.filter(idx, result.fids)]
        else:
            result[idx] = else_[else_.filter(idx, result.fids)]
    return result


@_evaluate.register(gem.Indexed)
def _evaluate_indexed(e, self):
    """Indexing maps shape to free indices"""
    val = self(e.children[0])
    fids = tuple(i for i in e.multiindex if isinstance(i, gem.Index))

    idx = []
    # First pick up all the existing free indices
    for _ in val.fids:
        idx.append(slice(None))
    # Now grab the shape axes
    for i in e.multiindex:
        if isinstance(i, gem.Index):
            # Free index, want entire extent
            idx.append(slice(None))
        elif isinstance(i, gem.VariableIndex):
            # Variable index, evaluate inner expression
            result, = self(i.expression)
            assert not result.tshape
            idx.append(result[()])
        else:
            # Fixed index, just pick that value
            idx.append(i)
    assert len(idx) == len(val.tshape)
    return Result(val[idx], val.fids + fids)


@_evaluate.register(gem.ComponentTensor)
def _evaluate_componenttensor(e, self):
    """Component tensors map free indices to shape."""
    val = self(e.children[0])
    axes = []
    fids = []
    # First grab the free indices that aren't bound
    for a, f in enumerate(val.fids):
        if f not in e.multiindex:
            axes.append(a)
            fids.append(f)
    # Now the bound free indices
    for i in e.multiindex:
        axes.append(val.fids.index(i))
    # Now the existing shape
    axes.extend(range(len(val.fshape), len(val.tshape)))
    return Result(numpy.transpose(val.arr, axes=axes),
                  tuple(fids))


@_evaluate.register(gem.IndexSum)
def _evaluate_indexsum(e, self):
    """Index sums reduce over the given axis."""
    val = self(e.children[0])
    idx = tuple(map(val.fids.index, e.multiindex))
    rfids = tuple(fi for fi in val.fids if fi not in e.multiindex)
    return Result(val.arr.sum(axis=idx), rfids)


@_evaluate.register(gem.ListTensor)
def _evaluate_listtensor(e, self):
    """List tensors just turn into arrays."""
    ops = [self(o) for o in e.children]
    tmp = Result.empty(*ops)
    arrs = [numpy.broadcast_to(o.broadcast(tmp.fids), tmp.fshape) for o in ops]
    arrs = numpy.moveaxis(numpy.asarray(arrs), 0, -1).reshape(tmp.fshape + e.shape)
    return Result(arrs, tmp.fids)


@_evaluate.register(gem.Concatenate)
def _evaluate_concatenate(e, self):
    """Concatenate nodes flatten and concatenate shapes."""
    ops = [self(o) for o in e.children]
    fids = tuple(OrderedDict.fromkeys(itertools.chain(*(o.fids for o in ops))))
    fshape = tuple(i.extent for i in fids)
    arrs = []
    for o in ops:
        # Create temporary with correct shape
        arr = numpy.empty(fshape + o.shape)
        # Broadcast for extra free indices
        arr[:] = o.broadcast(fids)
        # Flatten shape
        arr = arr.reshape(arr.shape[:arr.ndim-len(o.shape)] + (-1,))
        arrs.append(arr)
    arrs = numpy.concatenate(arrs, axis=-1)
    return Result(arrs, fids)


def evaluate(expressions, bindings=None):
    """Evaluate some GEM expressions given variable bindings.

    :arg expressions: A single GEM expression, or iterable of
        expressions to evaluate.
    :kwarg bindings: An optional dict mapping GEM :class:`gem.Variable`
        nodes to data.
    :returns: a list of the evaluated expressions.
    """
    try:
        exprs = tuple(expressions)
    except TypeError:
        exprs = (expressions, )
    mapper = node.Memoizer(_evaluate)
    mapper.bindings = bindings if bindings is not None else {}
    return list(map(mapper, exprs))



import FIAT

from gem import ListTensor

from finat.fiat_elements import FiatElement
from finat.physically_mapped import Citations, identity, PhysicallyMappedElement
from finat.piola_mapped import normal_tangential_edge_transform
from copy import deepcopy


class MardalTaiWinther(PhysicallyMappedElement, FiatElement):
    def __init__(self, cell, degree=3):
        if Citations is not None:
            Citations().register("Mardal2002")
        super().__init__(FIAT.MardalTaiWinther(cell, degree))

        reduced_dofs = deepcopy(self._element.entity_dofs())
        sd = cell.get_spatial_dimension()
        fdofs = sd + 1
        reduced_dofs[sd][0] = []
        for f in reduced_dofs[sd-1]:
            reduced_dofs[sd-1][f] = reduced_dofs[sd-1][f][:fdofs]
        self._entity_dofs = reduced_dofs
        self._space_dimension = fdofs * len(reduced_dofs[sd-1])

    def basis_transformation(self, coordinate_mapping):
        numbf = self._element.space_dimension()
        ndof = self.space_dimension()
        V = identity(numbf, ndof)

        sd = self.cell.get_spatial_dimension()
        bary, = self.cell.make_points(sd, 0, sd+1)
        J = coordinate_mapping.jacobian_at(bary)
        detJ = coordinate_mapping.detJ_at(bary)
        entity_dofs = self.entity_dofs()
        for f in sorted(entity_dofs[sd-1]):
            cur = entity_dofs[sd-1][f][0]
            V[cur+1, cur:cur+sd] = normal_tangential_edge_transform(self.cell, J, detJ, f)

        return ListTensor(V.T)

    def entity_dofs(self):
        return self._entity_dofs

    def space_dimension(self):
        return self._space_dimension



"""Schedules operations to evaluate a multi-root expression DAG,
forming an ordered list of Impero terminals."""

import collections
import functools
import itertools

from gem import gem, impero
from gem.node import collect_refcount


class OrderedDefaultDict(collections.OrderedDict):
    """A dictionary that provides a default value and ordered iteration.

    :arg factory: The callable used to create the default value.

    See :class:`collections.OrderedDict` for description of the
    remaining arguments.
    """
    def __init__(self, factory, *args, **kwargs):
        self.factory = factory
        super(OrderedDefaultDict, self).__init__(*args, **kwargs)

    def __missing__(self, key):
        val = self[key] = self.factory()
        return val


class ReferenceStager(object):
    """Provides staging for nodes in reference counted expression
    DAGs.  A callback function is called once the reference count is
    exhausted."""

    def __init__(self, reference_count, callback):
        """Initialises a ReferenceStager.

        :arg reference_count: initial reference counts for all
                              expected nodes
        :arg callback: function to call on each node when
                       reference count is exhausted
        """
        self.waiting = reference_count.copy()
        self.callback = callback

    def decref(self, o):
        """Decreases the reference count of a node, and possibly
        triggering a callback (when the reference count drops to
        zero)."""
        assert 1 <= self.waiting[o]

        self.waiting[o] -= 1
        if self.waiting[o] == 0:
            self.callback(o)

    def empty(self):
        """All reference counts exhausted?"""
        return not any(self.waiting.values())


class Queue(object):
    """Special queue for operation scheduling.  GEM / Impero nodes are
    inserted when they are ready to be scheduled, i.e. any operation
    which depends on the operation to be inserted must have been
    scheduled already.  This class implements a heuristic for ordering
    operations within the constraints in a way which aims to achieve
    maximum loop fusion to minimise the size of temporaries which need
    to be introduced.
    """
    def __init__(self, callback):
        """Initialises a Queue.

        :arg callback: function called on each element "popped" from the queue
        """
        # Must have deterministic iteration over the queue
        self.queue = OrderedDefaultDict(list)
        self.callback = callback

    def insert(self, indices, elem):
        """Insert element into queue.

        :arg indices: loop indices used by the scheduling heuristic
        :arg elem: element to be scheduled
        """
        self.queue[indices].append(elem)

    def process(self):
        """Pops elements from the queue and calls the callback
        function on them until the queue is empty.  The callback
        function can insert further elements into the queue.
        """
        indices = ()
        while self.queue:
            # Find innermost non-empty outer loop
            while indices not in (i[:len(indices)] for i in self.queue.keys()):
                indices = indices[:-1]

            # Pick a loop
            for i in self.queue.keys():
                if i[:len(indices)] == indices:
                    indices = i
                    break

            while self.queue[indices]:
                self.callback(self.queue[indices].pop())
            del self.queue[indices]


def handle(ops, push, decref, node):
    """Helper function for scheduling"""
    if isinstance(node, gem.Variable):
        # Declared in the kernel header
        pass
    elif isinstance(node, gem.Constant):
        # Constant literals inlined, unless tensor-valued
        if node.shape:
            ops.append(impero.Evaluate(node))
    elif isinstance(node, gem.Zero):  # should rarely happen
        assert not node.shape
    elif isinstance(node, (gem.Indexed, gem.FlexiblyIndexed)):
        if node.indirect_children:
            # Do not inline;
            # Index expression can be involved if it contains VariableIndex.
            ops.append(impero.Evaluate(node))
        for child in itertools.chain(node.children, node.indirect_children):
            decref(child)
    elif isinstance(node, gem.IndexSum):
        ops.append(impero.Noop(node))
        push(impero.Accumulate(node))
    elif isinstance(node, gem.Node):
        ops.append(impero.Evaluate(node))
        for child in node.children:
            decref(child)
    elif isinstance(node, impero.Initialise):
        ops.append(node)
    elif isinstance(node, impero.Accumulate):
        ops.append(node)
        push(impero.Initialise(node.indexsum))
        decref(node.indexsum.children[0])
    elif isinstance(node, impero.Return):
        ops.append(node)
        decref(node.expression)
    elif isinstance(node, impero.ReturnAccumulate):
        ops.append(node)
        decref(node.indexsum.children[0])
    else:
        raise AssertionError("no handler for node type %s" % type(node))


def emit_operations(assignments, get_indices, emit_return_accumulate=True):
    """Makes an ordering of operations to evaluate a multi-root
    expression DAG.

    :arg assignments: Iterable of (variable, expression) pairs.
                      The value of expression is written into variable
                      upon execution.
    :arg get_indices: mapping from GEM nodes to an ordering of free
                      indices
    :arg emit_return_accumulate: emit ReturnAccumulate nodes? Set to
                      False if the output variables are not guaranteed
                      zero on entry to the kernel.
    :returns: list of Impero terminals correctly ordered to evaluate
              the assignments
    """
    # Prepare reference counts
    refcount = collect_refcount([e for v, e in assignments])

    # Stage return operations
    staging = []
    for variable, expression in assignments:
        if emit_return_accumulate and \
                refcount[expression] == 1 and isinstance(expression, gem.IndexSum) \
                and set(variable.free_indices) == set(expression.free_indices):
            staging.append(impero.ReturnAccumulate(variable, expression))
            refcount[expression] -= 1
        else:
            staging.append(impero.Return(variable, expression))

    # Prepare data structures
    def push_node(node):
        queue.insert(get_indices(node), node)

    def push_op(op):
        queue.insert(op.loop_shape(get_indices), op)

    ops = []

    stager = ReferenceStager(refcount, push_node)
    queue = Queue(functools.partial(handle, ops, push_op, stager.decref))

    # Enqueue return operations
    for op in staging:
        push_op(op)

    # Schedule operations
    queue.process()

    # Assert that nothing left unprocessed
    assert stager.empty()

    # Return
    ops.reverse()
    return ops
